import cterasdk.settings
from ...clients import clients
from ..services import Management
from ..endpoints import EndpointBuilder
from .. import authenticators
from ...lib.session.edge import Session

from ...edge import (
    afp, aio, array, audit, backup, cache, cli, config, connection, ctera_migrate,
    dedup, directoryservice, drive, files, firmware, ftp, groups, licenses, login,
    logs, mail, network, nfs, ntp, power, remote, rsync, ransom_protect, services,
    shares, shell, smb, snmp, ssh, ssl, support, sync, syslog, taskmgr, telnet,
    timezone, users, volumes,
)


class Clients:

    def __init__(self, edge, Portal):
        if Portal:
            edge._Portal = Portal
            edge.default.close()
            edge._ctera_session.start_remote_session(Portal.session())
            self.api = Portal.default.clone(clients.API, EndpointBuilder.new(edge.base), authenticator=lambda *_: True)
        else:
            self.migrate = edge.default.clone(clients.Migrate, EndpointBuilder.new(edge.base, '/migration/rest/v1'))
            self.api = edge.default.clone(clients.API, EndpointBuilder.new(edge.base, '/admingui/api'))
            self.io = IO(edge)


class IO:

    def __init__(self, edge):
        self._edge = edge
        self._webdav = edge.default.clone(clients.WebDAV, EndpointBuilder.new(edge.base, '/localFiles'))

    @property
    def download(self):
        return self._webdav.download

    @property
    def download_zip(self):
        return self._edge.default.form_data  # pylint: disable=protected-access

    @property
    def upload(self):
        return self._edge.default.form_data  # pylint: disable=protected-access

    @property
    def propfind(self):
        return self._webdav.propfind

    @property
    def mkdir(self):
        return self._webdav.mkcol

    @property
    def copy(self):
        return self._webdav.copy

    @property
    def move(self):
        return self._webdav.move

    @property
    def delete(self):
        return self._webdav.delete


class Edge(Management):  # pylint: disable=too-many-instance-attributes

    def __init__(self, host=None, port=None, https=True, Portal=None, *, base=None):
        super().__init__(host, port, https, base, cterasdk.settings.edge.syn.settings)
        self._ctera_session = Session(self.host())
        self._ctera_clients = Clients(self, Portal)
        self.afp = afp.AFP(self)
        self.aio = aio.AIO(self)
        self.array = array.Array(self)
        self.audit = audit.Audit(self)
        self.backup = backup.Backup(self)
        self.cache = cache.Cache(self)
        self.cli = cli.CLI(self)
        self.config = config.Config(self)
        self.ctera_migrate = ctera_migrate.CTERAMigrate(self)
        self.dedup = dedup.Dedup(self)
        self.directoryservice = directoryservice.DirectoryService(self)
        self.drive = drive.Drive(self)
        self.files = files.FileBrowser(self)
        self.firmware = firmware.Firmware(self)
        self.ftp = ftp.FTP(self)
        self.groups = groups.Groups(self)
        self.licenses = licenses.Licenses(self)
        self.logs = logs.Logs(self)
        self.mail = mail.Mail(self)
        self.network = network.Network(self)
        self.nfs = nfs.NFS(self)
        self.ntp = ntp.NTP(self)
        self.power = power.Power(self)
        self.ransom_protect = ransom_protect.RansomProtect(self)
        self.rsync = rsync.RSync(self)
        self.services = services.Services(self)
        self.shares = shares.Shares(self)
        self.shell = shell.Shell(self)
        self.smb = smb.SMB(self)
        self.snmp = snmp.SNMP(self)
        self.ssh = ssh.SSH(self)
        self.ssl = ssl.SSL(self)
        self.support = support.Support(self)
        self.sync = sync.Sync(self)
        self.syslog = syslog.Syslog(self)
        self.tasks = taskmgr.Tasks(self)
        self.telnet = telnet.Telnet(self)
        self.timezone = timezone.Timezone(self)
        self.users = users.Users(self)
        self.volumes = volumes.Volumes(self)

    def _after_login(self):
        self.ssl = ssl.initialize(self)

    @property
    def migrate(self):
        return self.clients.migrate

    @property
    def api(self):
        return self.clients.api

    @property
    def io(self):
        return self.clients.io

    @property
    def _session_id_key(self):
        return '_cteraSessionId_'

    def _authenticator(self, url):
        return authenticators.edge(self.session(), url)

    @property
    def _login_object(self):
        return login.Login(self)

    @property
    def initialized(self):
        return not self._login_object.info().isfirstlogin

    def test(self):
        return connection.test(self)

    def sso(self, ticket):
        """ Login using Single Sign On"""
        self._login_object.sso(ticket)
        self.session().start_session(self)

    def remote_access(self):
        return remote.remote_access(self, self._Portal)

    @property
    def _omit_fields(self):
        return super()._omit_fields + ['afp', 'aio', 'array', 'audit', 'backup', 'cache', 'cli', 'config', 'ctera_migrate', 'dedup',
                                       'directoryservice', 'drive', 'files', 'firmware', 'ftp', 'groups', 'licenses', 'logs', 'mail',
                                       'network', 'nfs', 'ntp', 'power', 'ransom_protect', 'rsync', 'services', 'shares', 'shell',
                                       'smb', 'snmp', 'ssh', 'ssl', 'support', 'sync', 'syslog', 'tasks', 'telnet', 'timezone',
                                       'users', 'volumes']



from .iterator import QueryAsyncIterator
from ...core.query import QueryParams
from ...lib import DefaultResponse, Command


def create_callback_function(core, path, name=None, *, callback_response=None):
    """
    Create a query callback function

    :param cterasdk.objects.core.Portal core: Portal object
    :param cterasdk.core.query.QueryParams param: Query paramter object
    :param str,optional name: Schema method name
    :param cterasdk.lib.iterator.BaseResponse callback_response: Class to consume callback response

    :returns: Command object
    """

    async def database(core, path, name, param):
        return callback_response(await core.v1.api.database(path, name, param))

    async def execute(core, path, name, param):
        return callback_response(await core.v1.api.execute(path, name, param))

    return Command(execute if name else database, core, path, name or 'query')


def iterator(core, path, param=None, name=None, *, callback_response=None):
    """
    Create iterator

    :param cterasdk.objects.core.Portal core: Portal object
    :param str path: URL Path
    :param str,optional name: Schema method name
    :param cterasdk.core.query.QueryParams,optional param: Query paramter object
    :param cterasdk.lib.iterator.BaseResponse callback_response: Class to consume callback response

    :returns: Query iterator object
    """

    callback_response = callback_response if callback_response else DefaultResponse
    callback_function = create_callback_function(core, path, name, callback_response=callback_response)
    return QueryAsyncIterator(callback_function, param if param else QueryParams())



import logging

from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.core')


class Activation(BaseCommand):
    """ Portal activation """

    def generate_code(self, username=None, tenant=None):
        """
        Generate device activation code

        :param str,optional username: User name used for activation, defaults to None
        :param str,optional tenant: Tenant name used for activation, defaults to None
        :return: Portal Activation Code
        :rtype: str
        """
        params = None
        if username:
            params = {'username': username}
        if tenant:
            params['portal'] = tenant

        logger.info('Generating device activation code. %s', {'user': username, 'portal': tenant})
        response = self._core.api.get('/ssoActivation', params=params)
        logger.info('Generated device activation code. %s', {'user': username, 'portal': tenant})

        return response.code



from ..exceptions import UserConsentError


def ask(question):
    answer = None
    answers = ['Y', 'n']
    question = question + ' (' + '/'.join(answers) + '): '

    try:
        while answer not in answers:
            try:
                answer = input(question)
            except EOFError:
                raise UserConsentError()

    except KeyboardInterrupt:
        answer = 'n'
    if answer == 'Y':
        return True
    return False



import re
import uuid
import atexit
import logging
import cterasdk.settings
from ..lib.storage import synfs, commonfs
from ..common import Object, utf8_decode, utf8_encode
from ..convert import fromjsonstr, fromxmlstr, tojsonstr, toxmlstr
from ..objects import uri


class Collection(Object):
    """Postman Collection"""

    __instance = None

    def __init__(self):
        super().__init__()
        self.info = Info()
        self.item = []
        Collection.__instance = self

    @staticmethod
    def instance():
        if Collection.__instance is None:
            Collection()
        return Collection.__instance

    def add(self, request):
        name = f'{request.method} /{"/".join(request.url.path)}'
        self.item.append(Command(name, request))

    def serialize(self):
        return str(self)


class Info(Object):

    def __init__(self):
        super().__init__()
        self.name = f'{str(uuid.uuid4())}'
        self.schema = "https://schema.getpostman.com/json/collection/v2.1.0/collection.json"


class Command(Object):

    def __init__(self, name, request):
        super().__init__()
        self.name = name
        self.request = request
        self.response = []


class Request(Object):

    def __init__(self, method, url):
        super().__init__()
        self.method = method
        self.header = None
        self.url = url

    def request_headers(self, d):
        self.header = [Header(k, v) for k, v in d.items()]

    def request_body(self, data):
        self.body = data  # pylint: disable=attribute-defined-outside-init


class Header(Object):

    def __init__(self, key, value):
        super().__init__()
        self.key = key
        self.value = value
        self.type = 'text'


class BodyStream:

    def __init__(self):
        self.data = bytearray()
        self.maxsize = 4194304  # 4 MB

    def append(self, data):
        if len(self.data) >= self.maxsize:
            return
        self.data = self.data + data[:self.maxsize - len(self.data)]

    def deserialize(self, mime_type):
        if self.data:
            if mime_type is not None:
                if mime_type == 'application/x-www-form-urlencoded':
                    form_data = uri.parse_qsl(utf8_decode(self.data))
                    if form_data:
                        form = Form()
                        for k, v in form_data:
                            form.add(k, v)
                        return form
                elif mime_type == 'text/plain':
                    return Raw.xml(toxmlstr(fromxmlstr(self.data), no_log=True))
                elif mime_type == 'application/json':
                    return Raw.json(tojsonstr(fromjsonstr(self.data), False))
                elif mime_type.startswith('multipart/form-data'):
                    return form_data_generator(self.data, mime_type)
        return None


def form_data_generator(data, mime_type):
    """
    Generate Form Data.

    :param bytearray data: Body.
    :param str mime_type: Header.
    """
    boundary = form_boundary(mime_type)
    keys = form_keys(data)
    if keys:
        form_data = FormData()
        segments = data.split(boundary)
        for index, key in enumerate(keys):
            value = form_value(segments[index + 1], key)
            form_data.add(key, value)
        return form_data
    return None


def form_value(data, key):
    """
    Get Value for Key.

    :param bytearray data: Segment of Body.
    :param bytes key: UTF-8 Encoded Key.
    """
    CRLF = '\r\n\r\n'
    expr = utf8_encode(f'(?<={key}"{CRLF})[^\r]+')
    match = re.search(expr, data, re.DOTALL)
    return match.group() if match else None


def form_keys(data):
    """
    Get Form Keys from Data.

    :param bytearray data: Body.
    :returns: UTF-8 Encoded Keys
    :rtype: list[bytes]
    """
    return [utf8_decode(key) for key in re.findall(rb'(?<=\ name=")[^"]+', data)]


def form_boundary(mime_type):
    """
    Get Multiparty Form Boundary from Content Type Header.

    :param str mime_type: Header.
    :returns: UTF-8 Encoded Boundary.
    :rtype: bytes
    """
    match = re.search(r'(?<=boundary=)[A-Za-z0-9]+', mime_type)
    return utf8_encode(f'--{match.group()}') if match else None


class Body(Object):
    """Request Body"""

    def __init__(self, mode):
        super().__init__()
        self.mode = mode


class Form(Body):

    def __init__(self):
        super().__init__('urlencoded')
        self.urlencoded = []

    def add(self, key, value):
        param = Object()
        param.key = key
        if key in ['j_password']:
            value = '*** Protected Value ***'
        param.value = value
        param.type = 'text'
        self.urlencoded.append(param)


class FormData(Body):

    def __init__(self):
        super().__init__('formdata')
        self.formdata = []

    def add(self, key, value):
        param = Object()
        param.key = key
        if key == 'file':
            param.type = 'file'
            param.src = ''
        else:
            param.type = 'text'
            param.value = utf8_decode(value)
        self.formdata.append(param)


class Raw(Body):

    def __init__(self, body, language):
        super().__init__('raw')
        self.raw = body if language == 'json' else utf8_decode(body)
        self.options = Object()
        self.options.raw = Object()
        self.options.language = language

    @staticmethod
    def xml(body):
        return Raw(body, 'xml')

    @staticmethod
    def json(body):
        return Raw(body, 'json')


class URL(Object):

    def __init__(self, raw, protocol, host, port, path, query):
        super().__init__()
        self.raw = raw
        self.protocol = protocol
        self.host = host
        self.port = port
        self.path = [uri.quote(part) for part in path]
        self.query = query


@atexit.register
def audit():
    if cterasdk.settings.audit.enabled:
        collection = Collection.instance()
        name = cterasdk.settings.audit.filename
        collection.info.name = name if name is not None else str(uuid.uuid4())
        filename = f'{collection.info.name}.json'
        logging.getLogger('cterasdk.http').info('Saving Postman audit file. %s', {
            'directory': commonfs.downloads(),
            'name': filename
        })
        synfs.write(commonfs.downloads(), filename, utf8_encode(collection.serialize()))



import logging
import gzip
import snappy
from ..exceptions.direct import DirectIOError


logger = logging.getLogger('cterasdk.direct')


def decompress(compressed_block):
    """
    Decompress a Block.

    :param bytes block: Compressed Block
    :returns: Decompressed Block
    :rtype: bytes
    """
    if compressed_block.startswith(b'\x1f\x8b'):  # Gzip Standard Declaration.
        try:
            return gzip.decompress(compressed_block)
        except gzip.BadGzipFile:
            logger.error('Failed to Decompress Block. Bad Gzip.')
            raise DirectIOError()

    try:
        return decompress_with_magic_header(compressed_block) \
            if compressed_block.startswith(b'\x82SNAPPY\x00') else snappy.uncompress(compressed_block)  # Snappy Magic
    except snappy.UncompressError:
        logger.error('Failed to Decompress Block.')
        raise DirectIOError()


def decompress_with_magic_header(compressed_block):
    """
    Decompress a Block.

    :param bytes block: Compressed Block
    :returns: Decompressed Block
    :rtype: bytes
    """
    logger.debug('Decompressing Block.')
    decompressed_block = bytearray()
    size_of_data = len(compressed_block)
    chunk_size, chunk_start = 4, 16
    while chunk_start < size_of_data:
        chunk_end = chunk_start + chunk_size + int.from_bytes(compressed_block[chunk_start:chunk_start + chunk_size], byteorder='big')
        chunk_start = chunk_start + chunk_size
        if chunk_end > size_of_data:
            break
        decompressed_block.extend(snappy.decompress(compressed_block[chunk_start:chunk_end]))
        chunk_start = chunk_end
    return decompressed_block



from http import HTTPStatus
from .base import CTERAException


class HTTPError(CTERAException):
    """
    HTTP Error

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, status, error):
        super().__init__(error.request.url)
        self.code = status.value
        self.name = status.name
        self.error = error


class BadRequest(HTTPError):
    """
    Bad Request

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, error):
        super().__init__(HTTPStatus.BAD_REQUEST, error)


class Unauthorized(HTTPError):
    """
    Unauthorized

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, error):
        super().__init__(HTTPStatus.UNAUTHORIZED, error)


class Forbidden(HTTPError):
    """
    Unauthorized

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, error):
        super().__init__(HTTPStatus.FORBIDDEN, error)


class NotFound(HTTPError):
    """
    NotFound

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, error):
        super().__init__(HTTPStatus.NOT_FOUND, error)


class Unprocessable(HTTPError):
    """
    Unprocessable

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, error):
        super().__init__(HTTPStatus.UNPROCESSABLE_ENTITY, error)


class InternalServerError(HTTPError):
    """
    InternalServerError

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, error):
        super().__init__(HTTPStatus.INTERNAL_SERVER_ERROR, error)


class BadGateway(HTTPError):
    """
    BadGateway

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, error):
        super().__init__(HTTPStatus.BAD_GATEWAY, error)


class ServiceUnavailable(HTTPError):
    """
    ServiceUnavailable

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, error):
        super().__init__(HTTPStatus.SERVICE_UNAVAILABLE, error)


class GatewayTimeout(HTTPError):
    """
    GatewayTimeout

    :ivar int code: Status code
    :ivar str name: Reason
    :ivar cterasdk.clients.errors.Error error: Error object
    """

    def __init__(self, error):
        super().__init__(HTTPStatus.GATEWAY_TIMEOUT, error)



class DayOfWeek:
    """
    Day of Week

    :ivar int Sunday: Sunday
    :ivar int Monday: Monday
    :ivar int Tuesday: Tuesday
    :ivar int Wednesday: Wednesday
    :ivar int Thursday: Thursday
    :ivar int Friday: Friday
    :ivar int Saturday: Saturday
    """
    Sunday = 0
    Monday = 1
    Tuesday = 2
    Wednesday = 3
    Thursday = 4
    Friday = 5
    Saturday = 6
    Weekdays = [Monday, Tuesday, Wednesday, Thursday, Friday]
    Weekend = [Saturday, Sunday]
    All = [Sunday, Monday, Tuesday, Wednesday, Thursday, Friday, Saturday]


class FileCriteria:
    """
    File Criteria

    :ivar str Name: File name
    :ivar str Path: File path
    :ivar str Type: File type
    :ivar str Size: File size
    :ivar str Modified: Last modified
    """
    Name = 'FileName'
    Path = 'PathName'
    Type = 'FileType'
    Size = 'FileSize'
    Modified = 'LastModified'


class BooleanFunction:
    """
    Boolean function

    :ivar str AND: AND
    :ivar str AND: OR
    """
    AND = 'AND'
    OR = 'OR'


class Application:
    """
    Application

    :ivar str All: All Applications
    :ivar str Exchange: Microsoft Exchange
    :ivar str SQL: Microsoft SQL Server
    :ivar str NTDS: Active Directory Users and Groups
    :ivar str DFS: DFS Replication Service
    :ivar str FRS: File Replication Service
    :ivar str System: System State
    :ivar str HyperV: Hyper-V
    :ivar str SharePoint_07: SharePoint 2007
    :ivar str SharePoint: SharePoint
    """
    Exchange = '{76fe1ac4-15f7-4bcd-987e-8e1acb462fb7}'
    SQL = '{a65faa63-5ea8-4ebc-9dbd-a0c4db26912a}'
    NTDS = '{b2014c9e-8711-4c5c-a5a9-3cf384484757}'
    DFS = '{2707761b-2324-473d-88eb-eb007a359533}'
    FRS = '{d76f5a28-3092-4589-ba48-2958fb88ce29}'
    System = '{db7f098c-121b-46fe-814b-58902de3a193}'
    HyperV = '{66841cd4-6ded-4f4b-8f17-fd23f8ddc3de}'
    SharePoint_07 = '{c2f52614-5e53-4858-a589-38eeb25c6184}'
    SharePoint = '{da452614-4858-5e53-a512-38aab25c61ad}'
    All = [Exchange, SQL, NTDS, DFS, FRS, System, HyperV, SharePoint_07, SharePoint]


class ScheduleType:
    """
    Schedule Type

    :ivar str Manual: Manual
    :ivar str Hourly: Hourly
    :ivar str Daily: Daily
    :ivar str Weekly: Weekly
    :ivar str Monthly: Monthly
    :ivar str Interval: Interval
    :ivar str Window: Window
    """
    Manual = 'manual'
    Hourly = 'hourly'
    Daily = 'daily'
    Weekly = 'weekly'
    Monthly = 'monthly'
    Interval = 'interval'
    Window = 'window'



import logging

from ..exceptions import CTERAException
from .enum import Mode, IPProtocol, Traffic
from .types import TCPConnectResult
from ..lib.task_manager_base import TaskError
from ..common import Object, parse_to_ipaddress
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class Network(BaseCommand):
    """ Edge Filer Network configuration APIs """

    def __init__(self, portal):
        super().__init__(portal)
        self.proxy = Proxy(self._edge)
        self.mtu = MTU(self._edge)
        self.routes = StaticRoutes(self._edge)

    def get_status(self):
        """
        Retrieve the network interface status
        """
        return self._edge.api.get('/status/network/ports/0')

    def ifconfig(self):
        """
        Retrieve the ip configuration
        """
        return self.ipconfig()

    def ipconfig(self):
        """
        Retrieve the ip configuration
        """
        return self._edge.api.get('/config/network/ports/0')

    def set_static_ipaddr(self, address, subnet, gateway, primary_dns_server, secondary_dns_server=None):
        """
        Set a Static IP Address

        :param str address: The static address
        :param str subnet: The subnet for the static address
        :param str gateway: The default gateway
        :param str primary_dns_server: The primary DNS server
        :param str,optinal secondary_dns_server: The secondary DNS server, defaults to None
        """
        ip = self._edge.api.get('/config/network/ports/0/ip')
        ip.DHCPMode = Mode.Disabled
        ip.address = address
        ip.netmask = subnet
        ip.gateway = gateway
        ip.autoObtainDNS = False
        ip.DNSServer1 = primary_dns_server

        if secondary_dns_server is not None:
            ip.DNSServer2 = secondary_dns_server

        logger.info('Configuring a static ip address.')

        self._edge.api.put('/config/network/ports/0/ip', ip)

        logger.info(
            'Network settings updated. %s',
            {'address': address, 'subnet': subnet, 'gateway': gateway, 'DNS1': primary_dns_server, 'DNS2': secondary_dns_server}
        )

    def set_static_nameserver(self, primary_dns_server, secondary_dns_server=None):
        """
        Set the DNS Server addresses statically

        :param str primary_dns_server: The primary DNS server
        :param str,optinal secondary_dns_server: The secondary DNS server, defaults to None
        """
        ip = self._edge.api.get('/config/network/ports/0/ip')
        ip.autoObtainDNS = False
        ip.DNSServer1 = primary_dns_server

        if secondary_dns_server is not None:
            ip.DNSServer2 = secondary_dns_server

        logger.info('Configuring nameserver settings.')

        self._edge.api.put('/config/network/ports/0/ip', ip)

        logger.info('Nameserver settings updated. %s', {'DNS1': primary_dns_server, 'DNS2': secondary_dns_server})

    def enable_dhcp(self):
        """
        Enable DHCP
        """
        ip = self._edge.api.get('/config/network/ports/0/ip')
        ip.DHCPMode = Mode.Enabled
        ip.autoObtainDNS = True

        logger.info('Enabling DHCP.')

        self._edge.api.put('/config/network/ports/0/ip', ip)

        logger.info('Network settings updated. Enabled DHCP.')

    def diagnose(self, services):
        """
        Test a TCP connection to a host over a designated port

        :param list[cterasdk.edge.types.TCPService] services: List of services, identified by a host and a port
        :returns: A list of named-tuples including the host, port and a boolean value indicating whether TCP connection can be established
        :rtype: list[cterasdk.edge.types.TCPConnectResult]
        """
        return [self.tcp_connect(service) for service in services]

    def tcp_connect(self, service):
        """
        Test a TCP connection between the Edge Filer and the provided host address

        :param cterasdk.edge.types.TCPService service: A service, identified by a host and a port
        :returns: A named-tuple including the host, port and a boolean value indicating whether TCP connection can be established
        :rtype: cterasdk.edge.types.TCPConnectResult
        """
        param = Object()
        param.address = service.host
        param.port = service.port

        logger.info("Testing connection. %s", {'host': service.host, 'port': service.port})

        task = self._edge.api.execute("/status/network", "tcpconnect", param, {
            'timeout': {
                'sock_read': 120
            }
        })
        try:
            task = self._edge.tasks.wait(task)
            logger.debug("Obtained connection status. %s", {'status': task.result.rc})
            if task.result.rc == "Open":
                return TCPConnectResult(service.host, service.port, True)
        except TaskError:
            pass

        logger.warning("Couldn't establish TCP connection. %s", {'address': service.host, 'port': service.port})

        return TCPConnectResult(service.host, service.port, False)

    def iperf(self, address, port=5201, threads=1, protocol=IPProtocol.TCP, direction=Traffic.Upload, retries=120, seconds=1):
        """
        Invoke a network throughput test

        :param str address: The host running the iperf server
        :param int,optional port: The iperf server port, defaults to 5201
        :param int,optional threads: The number of threads, defaults to 1
        :param cterasdk.edge.enum.IPProtocol,optional protocol: IP protocol, defaults to `'TCP'`
        :param cterasdk.edge.enum.Traffic,optional direction: Traffic direction, defaults to `'Upload'`
        :param int,optional retries: Number of retries when sampling the iperf task status, defaults to 120
        :param int,optional seconds: Number of seconds to wait between retries, defaults to 1
        :returns: A string containing the iperf output
        :rtype: str
        """
        param = Object()
        param._classname = 'IperfParam'  # pylint: disable=protected-access
        param.address = address
        param.port = port
        param.threads = threads
        param.reverse = direction == Traffic.Download
        param.protocol = None if protocol == IPProtocol.TCP else IPProtocol.UDP
        task = self._edge.api.execute("/status/network", "iperf", param)
        try:
            task = self._edge.tasks.wait(task, retries, seconds)
            return task.result.res
        except TaskError as error:
            return error.task.result.res


class Proxy(BaseCommand):
    """Edge Filer Proxy Configuration APIs"""

    def get_configuration(self):
        """
        Get Proxy Configuration
        """
        return self._edge.api.get('/config/network/proxy')

    def is_enabled(self):
        """
        Check if Proxy Configuration is Enabled

        :returns: ``True`` if a proxy server was configured and ``False`` otherwise.
        :rtype: bool
        """
        return self._edge.api.get('/config/network/proxy/configurationMode') != 'NoProxy'

    def modify(self, address, port=None, username=None, password=None):
        """
        Modify Proxy Configuration

        :param str address: Proxy address
        :param int,optional port: Proxy port, defaults to ``8080``
        :param str,optional username: Username
        :param str,optional password: Password
        :returns: Proxy settings
        :rtype: cterasdk.common.object.Object
        """
        return self._configure(True, address, port, username, password)

    def _configure(self, enabled, address=None, port=None, username=None, password=None):
        param = Object()
        param._classname = 'ProxySettings'  # pylint: disable=protected-access
        param.configurationMode = 'Manual' if enabled else 'NoProxy'
        if enabled:
            param.port = port if port else 8080
            if address:
                param.address = address
            if username:
                param.username = username
            if password:
                param.password = password
        logger.info('Updating Proxy Server Configuration.')
        response = self._edge.api.put('/config/network/proxy', param)
        logger.info('Updated Proxy Server Configuration.')
        return response

    def disable(self):
        """
        Disable Proxy

        :returns: Proxy settings
        :rtype: cterasdk.common.object.Object
        """
        logger.info('Disabling Proxy.')
        return self._configure(False)


class MTU(BaseCommand):
    """Edge Filer MTU Configuration APIs"""

    def reset(self):
        """
        Set the default maximum transmission unit (MTU) settings
        """
        return self._configure(False, 1500)

    def modify(self, mtu):
        """
        Set a custom network maximum transmission unit (MTU)

        :param int mtu: Maximum transmission unit
        """
        return self._configure(True, mtu)

    def _configure(self, jumbo, mtu):
        settings = self._edge.api.get('/config/network/ports/0/ethernet')
        settings.jumbo = jumbo
        settings.mtu = mtu
        logger.info('Configuring MTU. %s', {'MTU': mtu})
        return self._edge.api.put('/config/network/ports/0/ethernet', settings)


class StaticRoutes(BaseCommand):
    """Edge Filer Static Route Configuration APIs"""

    def get(self):
        """
        Get All Static Routes
        """
        return self._edge.api.get('/config/network/static_routes')

    def add(self, source_ip, destination_ip_mask):
        """
        Add a Static Route

        :param str source_ip: The source IP (192.168.15.55)
        :param str destination_ip_mask: The destination IP and CIDR block (10.5.0.1/32)
        """
        try:
            param = Object()
            param.GwIP = str(parse_to_ipaddress(source_ip))
            param.DestIpMask = str(parse_to_ipaddress(destination_ip_mask)).replace("/", "_")
            res = self._edge.api.add('/config/network/static_routes', param)
            logger.info(
                "Static route updated. %s", {'Source': param.GwIP, 'Destination': destination_ip_mask})
            return res
        except CTERAException as error:
            logger.error("Static route creation failed.")
            raise CTERAException('Static route creation failed') from error

    def remove(self, destination_ip_mask):
        """
        Remove a Static Route

        :param str destination_ip_mask: The destination IP and CIDR block (10.5.0.1/32)
        """
        try:
            dest_ip_mask = str(parse_to_ipaddress(destination_ip_mask)).replace("/", "_")
            response = self._edge.api.delete(f'/config/network/static_routes/{dest_ip_mask}')
            logger.info(
                "Static route deleted. %s", {'Destination': dest_ip_mask})
            return response
        except CTERAException as error:
            logger.error("Static route deletion failed.")
            raise CTERAException('Static route deletion failed') from error

    def clear(self):
        """
        Clear All Static routes
        """
        try:
            self._edge.api.execute('/config/network', 'cleanStaticRoutes')
            logger.info('Static routes were deleted successfully')
        except CTERAException as error:
            logger.error("Failed to clear static routes")
            raise CTERAException('Failed to clear static routes') from error



import datetime


class DateTimeUtils():
    @staticmethod
    def get_expiration_date(expiration):
        """
        Get a datetime.date representation of the expiration date

        :param variable,optional expiration:
            The expiration value.
            Pass datetime.date for a specific date, integer for days from now, or True for immediate (yesterday)
        :return datetime.date: datetime.date representation of the expiration date
        """
        if isinstance(expiration, bool):
            expiration_date = datetime.date.today() - datetime.timedelta(days=1)
        elif isinstance(expiration, int):
            expiration_date = datetime.date.today() + datetime.timedelta(days=expiration)
        elif isinstance(expiration, datetime.date):
            expiration_date = expiration
        return expiration_date  # pylint: disable=possibly-used-before-assignment


def from_iso_format(time):
    """
    Parse datetime object from ISO 8601 format

    :param str time: Timestamp
    :returns: Datetime object
    :rtype: datetime.datetime
    """
    return datetime.datetime.fromisoformat(time)



import os
import errno
import logging
import mimetypes
from pathlib import Path
from contextlib import contextmanager
import cterasdk.settings


logger = logging.getLogger('cterasdk.filesystem')


def exists(path):
    """
    Check if a file or a directory exists

    :param str path: Path
    :returns: ``True`` if exists, ``False`` otherwise.
    :rtype: bool
    """
    return Path(path).exists()


def expanduser(path):
    """
    Return a new path with expanded ~ and ~user constructs

    :param str path: Path
    :returns: Absolute Path.
    :rtype: Path object
    """
    return Path(path).expanduser()


def is_dir(path):
    """
    Check is a directory.

    :param str path: Path
    :returns: ``True`` if a directory, ``False`` otherwise.
    :rtype: bool
    """
    p = expanduser(path)
    return p.is_dir()


def join(*paths):
    """
    Join Path objects.

    :param list[Path] paths: Path objects
    :rtype: Path
    """
    return Path(*paths)


def properties(path):
    """
    File properties.

    :param str path: Path
    :returns: File name, size and type
    :rtype: dict
    :raises: FileNotFoundError
    """
    p = expanduser(path)

    if not p.exists() or not p.is_file():
        logger.error('File not found: %s', p.as_posix())
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), p.as_posix())

    return dict(
        name=p.name,
        size=str(p.stat().st_size),
        mimetype=mimetypes.guess_type(path)
    )


def downloads():
    """
    Get downloads directory.

    :returns: Directory path
    :rtype: Path object
    """
    directory = expanduser(cterasdk.settings.io.downloads)
    if not is_dir(directory):
        logger.error('Directory not found: %s', directory)
        raise FileNotFoundError(errno.ENOENT, 'Directory not found', directory)
    return directory


def determine_zip_archive_name(directory, objects):
    """
    Name the zip archive after the folder name, unless the directory contains only one object.

    :rtype: str
    """
    if len(objects) > 1:
        path = Path(directory)
    else:
        path = Path(objects[0])
    return f'{path.stem}.zip'


def new_version(filename, version):
    """
    Append version number to file name.

    :param str filename: File name
    :param int version: File version
    :returns: File name appended with a version number
    :rtype: str
    """
    idx = filename.rfind('.')
    extension = ''
    if idx > 0:
        name = filename[:idx]
        extension = filename[idx:]
    else:
        name = filename
    return f'{name} ({str(version)}){extension}'


def rename(source, new_name):
    """
    Rename a file or a directory.

    :param str source: Path
    :param str new_name: New name
    :returns: Path after rename
    :rtype: str
    """
    source = expanduser(source)
    if not source.exists():
        logger.error('File not found: %s', source.as_posix())
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), source.as_posix())
    destination = source.parent.joinpath(new_name)
    source.rename(destination)
    return destination.as_posix()


class ResultContext:
    """Context Manager Result Context"""

    def __init__(self):
        self._value = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        self._value = v


@contextmanager
def write_new_version(directory, name, *, ctx=None):
    """
    Context manager for writing a file without conflicts.

    :param str directory: Parent directory
    :param str name: File name
    :param ResultContext,optional ctx: Result Context
    :returns: Path
    :rtype: str
    :raises: FileNotFoundError
    """
    parent = expanduser(directory)
    if not parent.exists() or not is_dir(parent):
        logger.error('Directory not found: %s', parent.as_posix())
        raise FileNotFoundError(errno.ENOENT, 'Directory not found', directory)

    tempfile = parent.joinpath(f'{name}.ctera')
    yield tempfile

    origin, version = name, 0
    while True:
        try:
            rename(tempfile, name)
            break
        except (FileExistsError, IsADirectoryError):
            logger.debug('File exists: %s', parent.joinpath(name))
            version = version + 1
            name = new_version(origin, version)

    p = parent.joinpath(name)
    logger.info('Saved: %s', p.as_posix())

    if ctx and isinstance(ctx, ResultContext):
        ctx.value = p.as_posix()


def split_file_directory(location):
    """
    Split file and directory.

    :param str path: Path

    Returns:
    1. (parent directory, file name), if a file exists
    2. (parent directory, file name), if a directory exists
    3. (parent directory, file name), if the parent directory exists
    4. Raises ``FileNotFoundError`` if neither the object nor the parent directory exist
    """
    p = expanduser(location)
    if p.exists():
        if p.is_dir():
            filename = None
        else:
            filename = p.name
            p = p.parent
    elif p.parent.exists():
        filename = p.name
        p = p.parent
    else:
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), location)
    return str(p.resolve()), filename


def generate_file_destination(destination=None, default_name=None):
    """
    Compute destination file path.

    :param str location: Path to a file or a folder
    :param str default_name: Default file name, unless ``location`` already specifies a file path
    :returns: Tuple including the destination directory and file name
    :rtype: tuple(str, str)
    """
    parent = name = None
    if destination:
        parent, name = split_file_directory(destination)
    else:
        parent = downloads()

    if not name:
        name = default_name

    return (parent, name)


def determine_directory_and_filename(p, objects=None, destination=None, archive=False):
    """
    Determine location to save file.

    :param str p: Path.
    :param list[str],optional objects: List of files or folders
    :param str,optional destination: Destination
    :param bool,optional archive: Compressed archive
    :returns: Directory and file name
    :rtype: tuple[str]
    """
    directory, name = None, None
    if destination:
        directory, name = split_file_directory(destination)
    else:
        directory = downloads()

    if not name:
        normalized = Path(p)
        if archive:
            name = determine_zip_archive_name(p, objects)
        else:
            name = normalized.name
    return directory, name



class CTERAException(Exception):
    """
    Base Exception.

    :parm str message: Error message
    """
    def __init__(self, message=None):
        super().__init__(message)

    def __repr__(self):
        return str(self)



from .synchronous.core import GlobalAdmin, ServicesPortal  # noqa: E402, F401
from .synchronous.edge import Edge  # noqa: E402, F401
from .synchronous.drive import Drive  # noqa: E402, F401
from .asynchronous.core import AsyncGlobalAdmin, AsyncServicesPortal  # noqa: E402, F401
from .asynchronous.edge import AsyncEdge  # noqa: E402, F401



from .base_command import BaseCommand
from . import query
from ...core.query import QueryParamBuilder, FilterBuilder
from ...common.utils import union


class CloudFS(BaseCommand):
    """
    CloudFS APIs

    :ivar cterasdk.core.cloudfs.CloudDrives drives: Object holding Cloud Drive Folders APIs
    """

    def __init__(self, core):
        super().__init__(core)
        self.drives = CloudDrives(self._core)


class CloudDrives(BaseCommand):
    """ Cloud Drive Folder APIs """

    default = ['name']

    async def find(self, name, owner, include=None):
        """
        Find a Cloud Drive Folder

        :param str name: Name of the Cloud Drive Folder to find
        :param cterasdk.core.types.UserAccount owner: User account of the folder group owner
        :param list[str] include: List of metadata fields to include in the response

        :returns: A Cloud Drive Folder
        """

        user = await self._core.users.get(owner, ['uid'])
        include = union(include or [], CloudDrives.default)
        builder = QueryParamBuilder().include(include).ownedBy(user.uid)
        builder.addFilter(FilterBuilder('name').eq(name))
        param = builder.build()
        return query.iterator(self._core, '/cloudDrives', param)



import logging
from ....cio.common import encode_request_parameter
from ....cio import edge as fs
from ....exceptions.transport import NotFound
from ....exceptions.io import RestrictedPathError, ResourceNotFoundError, NotADirectory


logger = logging.getLogger('cterasdk.edge')


async def listdir(edge, path):
    return fs.format_listdir_response(path.reference.as_posix(), await edge.io.propfind(path.absolute, 1))


async def walk(edge, path):
    paths = [fs.EdgePath.instance('/', path)]
    while len(paths) > 0:
        path = paths.pop(0)
        entries = await listdir(edge, path)
        for e in entries:
            if e.is_dir:
                paths.append(fs.EdgePath.instance('/', e))
            yield e


async def exists(edge, path):
    present, *_ = await metadata(edge, path, suppress_error=True)
    return present


async def metadata(edge, path, suppress_error=False):
    """
    Get item metadata.

    :returns: A tuple indicating if a file exists, and its metadata
    """
    try:
        return True, fs.format_listdir_response(None, await edge.io.propfind(path.absolute, 0))[0]
    except NotFound as error:
        if not suppress_error:
            raise ResourceNotFoundError(path.absolute) from error
        return False, None


async def ensure_directory(edge, directory, suppress_error=False):
    present, resource = await metadata(edge, directory, suppress_error=True)
    if (not present or not resource.is_dir) and not suppress_error:
        raise NotADirectory(directory.absolute)
    return resource.is_dir if present else False, resource


async def mkdir(edge, path):
    with fs.makedir(path) as param:
        await edge.io.mkdir(param)
    return path.absolute


async def makedirs(edge, path):
    directories = path.parts
    for i in range(1, len(directories) + 1):
        path = fs.EdgePath(path.scope, '/'.join(directories[:i]))
        try:
            await mkdir(edge, path)
        except RestrictedPathError:
            logger.warning('Creating a folder in the specified location is forbidden: %s', path.reference.as_posix())


async def copy(edge, path, destination=None, overwrite=False):
    with fs.copy(path, destination) as (src, dst):
        await edge.io.copy(src, dst, overwrite=overwrite)


async def move(edge, path, destination=None, overwrite=False):
    with fs.move(path, destination) as (src, dst):
        await edge.io.move(src, dst, overwrite=overwrite)


async def remove(edge, *paths):
    for path in fs.delete_generator(*paths):
        await edge.io.delete(path.absolute)


def handle(path):
    """
    Create function to retrieve file handle.

    :param cterasdk.cio.edge.EdgePath path: Path to file.
    :returns: Callable function to retrieve file handle.
    :rtype: callable
    """
    async def wrapper(edge):
        """
        Get file handle.

        :param cterasdk.objects.synchronous.edge.Edge edge: Edge Filer object.
        """
        with fs.handle(path) as param:
            return await edge.io.download(param)
    return wrapper


def handle_many(directory, *objects):
    """
    Create function to retrieve zip archive

    :param cterasdk.cio.edge.EdgePath directory: Path to directory.
    :param args objects: List of files and folders.
    :returns: Callable function to retrieve file handle.
    :rtype: callable
    """
    async def wrapper(edge):
        """
        Upload file from metadata and file handle.

        :param cterasdk.objects.synchronous.edge.Edge edge: Edge Filer object.
        :param str name: File name.
        :param object handle: File handle.
        """
        with fs.handle_many(directory, objects) as param:
            return await edge.io.download_zip('/admingui/api/status/fileManager/zip', encode_request_parameter(param))
    return wrapper


async def _validate_destination(edge, name, destination):
    is_dir, *_ = await ensure_directory(edge, destination, suppress_error=True)
    if not is_dir:
        is_dir, *_ = await ensure_directory(edge, destination.parent)
        return destination.name, destination.parent
    return name, destination


def upload(name, destination, fd):
    """
    Create upload function

    :param str name: File name.
    :param cterasdk.cio.edge.EdgePath destination: Path to directory.
    :param object fd: File handle.
    :returns: Callable function to start the upload.
    :rtype: callable
    """
    async def wrapper(edge):
        """
        Upload file from metadata and file handle.

        :param cterasdk.objects.synchronous.edge.Edge edge: Edge Filer object.
        """
        filename, directory = await _validate_destination(edge, name, destination)
        with fs.upload(filename, directory, fd) as param:
            return await edge.io.upload('/actions/upload', param)
    return wrapper



import logging
from .commonfs import write_new_version, ResultContext


logger = logging.getLogger('cterasdk.filesystem')


def write(directory, name, handle):
    """
    Write bytes to disk.

    :param str directory: Directory
    :param str name: Name
    :param bytes handle: Handle
    :returns: Path
    :rtype: str
    """
    ctx = ResultContext()
    with write_new_version(directory, name, ctx=ctx) as tempfile:
        overwrite(tempfile, handle)
    return ctx.value


def overwrite(p, handle):
    """
    Write, without validation.

    :param Path p: Path
    :param bytes handle: Handle
    """
    with open(p, 'w+b') as fd:
        if isinstance(handle, bytes):
            fd.write(handle)
        else:
            for chunk in handle.iter_content(chunk_size=8192):
                fd.write(chunk)
        logger.debug('Wrote: %s', p.as_posix())
    return p.as_posix()



from pathlib import PurePosixPath
from ..objects.uri import quote
from ..common.utils import utf8_decode
from ..convert.serializers import toxmlstr


class BasePath:
    """Base Path for CTERA Portal and CTERA Edge"""

    def __init__(self, scope, reference):
        """
        Initialize a Base Path.

        :param str scope: Scope.
        :param str reference: Reference.
        """
        if isinstance(scope, str):
            scope = scope.lstrip('/')
        if isinstance(reference, str):
            reference = reference.lstrip('/')
        self._scope = PurePosixPath(scope)
        self._reference = PurePosixPath(reference)

    @property
    def scope(self):
        return self._scope

    @property
    def reference(self):
        return self._reference

    @property
    def name(self):
        return self._reference.name

    @property
    def parent(self):
        return self.__class__(self._scope.as_posix(), self._reference.parent.as_posix())

    @property
    def absolute(self):
        return f'/{self.scope.joinpath(self.reference).as_posix()}'

    @property
    def absolute_encode(self):
        return f'/{self.scope.joinpath(quote(self.reference.as_posix())).as_posix()}'

    @property
    def absolute_parent(self):
        return self.parent.as_posix()

    def join(self, p):
        """
        Join Path.

        :param str p: Path.
        """
        return self.__class__(self._scope.as_posix(), self.reference.joinpath(p).as_posix())

    @property
    def parts(self):
        return self.reference.parts

    def __str__(self):
        return self.absolute


def encode_stream(fd, size):
    if isinstance(fd, str):
        fd = fd.encode('utf-8')
    if isinstance(fd, bytes):
        size = str(len(fd))
    return fd, size


def encode_request_parameter(param):
    return dict(
        inputXML=utf8_decode(toxmlstr(param))
    )



import logging

from ..common import Object
from ..exceptions import CTERAException
from ..exceptions.backup import NotFound, AttachEncrypted, IncorrectPassphrase, ClocksOutOfSync
from .enum import BackupConfStatusID
from .base_command import BaseCommand
from .directorytree import DirectoryTree


logger = logging.getLogger('cterasdk.edge')


class AttachRC:
    OK = 'OK'
    NotFound = 'NotFound'
    IsEncrypted = 'IsEncrypted'
    CheckCodeInCorrect = 'CheckCodeInCorrect'
    ClocksOutOfSync = 'ClocksOutOfSync'
    InternalServerError = 'InternalServerError'
    PermissionDenied = 'PermissionDenied'


class CreateFolderRC:
    OK = 'OK'
    InternalServerError = 'InternalServerError'
    PermissionDenied = 'PermissionDenied'
    FolderAlreadyExists = 'FolderAlreadyExists'


class EncryptionMode:
    """
    Encryption mode types

    :ivar str Recoverable: Recoverable key encryption mode
    :ivar str Secret: Secret key encryption mode
    """
    Recoverable = 'RecoverableKeyEncryption'
    Secret = 'SecretKeyEncryption'


class Backup(BaseCommand):
    """ Edge Filer backup configuration APIs """

    def __init__(self, edge):
        super().__init__(edge)
        self.files = BackupFiles(self._edge)

    def configure(self, passphrase=None):
        """
        Edge Filer backup configuration

        :param str,optional passphrase: Passphrase for the backup, defaults to None
        """
        logger.info('Configuring cloud backup.')

        try:
            settings = self._attach(passphrase)
        except NotFound:
            settings = self._create_folder(passphrase)

        self._configure_backup_settings(settings)

        logger.info('Cloud backup configuration completed successfully.')

    def is_configured(self):
        """
        Is Backup configured

        :return bool: True if backup is configured, else False
        """
        backup_status = self._edge.api.get('/proc/backup/backupStatus')
        return backup_status.serviceStatus.id == BackupConfStatusID.Attached

    def start(self):
        """ Start backup """
        logger.info("Starting cloud backup.")
        self._edge.api.execute("/status/sync", "start")

    def suspend(self):
        """ Suspend backup """
        logger.info("Suspending cloud backup.")
        self._edge.api.execute("/status/sync", "pause")

    def unsuspend(self):
        """ Unsuspend backup """
        logger.info("Suspending cloud backup.")
        self._edge.api.execute("/status/sync", "resume")

    def _attach(self, sharedSecret):
        try:
            logger.debug('Attaching to a backup folder.')
            settings = self._attach_folder()
        except AttachEncrypted as param:
            logger.debug('Attaching to an encrypted backup folder.')
            settings = self._attach_encrypted_folder(param.encryptedFolderKey, param.passPhraseSalt, sharedSecret)
            settings.encryptionMode = param.encryptionMode
        logger.debug('Successfully attached to a backup folder.')

        return settings

    def _attach_folder(self):
        task = self._edge.api.execute('/status/services', 'attachFolder')
        return self._attach_response(task)

    def _attach_encrypted_folder(self, encryptedFolderKey, passPhraseSalt, sharedSecret):
        param = Object()
        param.encryptedFolderKey = encryptedFolderKey
        param.passPhraseSalt = passPhraseSalt
        param.sharedSecret = sharedSecret

        task = self._edge.api.execute('/status/services', 'attachEncryptedFolder', param)
        return self._attach_response(task)

    def _attach_response(self, task):
        response = self._wait(task)
        return Backup._process_attach_response(response)

    @staticmethod
    def _process_attach_response(response):
        rc = response.attachFolderRC

        if rc == AttachRC.OK:
            param = Object()
            if hasattr(response, 'encryptionMode'):
                param.encryptionMode = response.encryptionMode
            param.sharedSecret = response.sharedSecret
            param.passPhraseSalt = response.passPhraseSalt
            return param

        if rc == AttachRC.NotFound:
            logger.debug('Could not find an existing backup folder.')
            raise NotFound()

        if rc == AttachRC.IsEncrypted:
            raise AttachEncrypted(response.encryptionMode, response.encryptedFolderKey, response.passPhraseSalt)

        if rc == AttachRC.CheckCodeInCorrect:
            logger.error('Incorrect passphrase.')
            raise IncorrectPassphrase()

        if rc == AttachRC.ClocksOutOfSync:
            logger.error('Intializing backup failed. Clocks are out of sync (rc=%s).', rc)
            raise ClocksOutOfSync()

        if rc == AttachRC.InternalServerError:
            logger.error('Attach failed (rc=%s).', rc)
        elif rc == AttachRC.PermissionDenied:
            logger.error('Attach failed (rc=%s).', rc)
        else:
            logger.error('Unknown error (rc=%s)', rc)
        raise CTERAException(f'Failed to attach to backup folder (rc={rc}).')

    def _create_folder(self, passphrase):
        param = Object()
        if passphrase is not None:
            logger.debug('Creting a passphrase-encrypted backup folder.')
            param.encryptionMode = EncryptionMode.Secret
            param.sharedSecret = passphrase
        else:
            logger.debug('Creating a backup folder.')
            param.encryptionMode = EncryptionMode.Recoverable

        task = self._edge.api.execute('/status/services', 'createFolder', param)
        settings = self._create_response(task)
        settings.encryptionMode = param.encryptionMode

        return settings

    def _create_response(self, task):
        response = self._wait(task)
        return Backup._process_create_response(response)

    @staticmethod
    def _process_create_response(response):
        rc = response.createFolderRC

        if rc == CreateFolderRC.OK:
            logger.debug('Backup folder created successfully.')
            param = Object()
            param.sharedSecret = response.sharedSecret
            param.passPhraseSalt = response.passPhraseSalt
            return param

        if rc == CreateFolderRC.InternalServerError:
            logger.error('Backup folder creation failed (rc=%s).', rc)
        elif rc == CreateFolderRC.PermissionDenied:
            logger.error('Backup folder creation failed (rc=%s).', rc)
        elif rc == CreateFolderRC.FolderAlreadyExists:
            return None
        raise CTERAException(f'Failed to create backup folder (rc={rc})')

    def _wait(self, task):
        task = self._edge.tasks.wait(task)
        return task.result

    def _configure_backup_settings(self, param):
        backup_settings = self._edge.api.get('/config/backup')
        if not backup_settings:
            backup_settings = self._edge.api.defaults('BackupSettings')

        backup_settings.encryptionMode = param.encryptionMode
        backup_settings.sharedSecret = param.sharedSecret
        backup_settings.passPhraseSalt = param.passPhraseSalt

        logger.debug('Configuring backup settings.')

        self._edge.api.put('/config/backup', backup_settings)


class BackupFiles(BaseCommand):

    ALL_FILES = 'All File Types'

    def unselect_all(self):
        """ Unselect all files from backup """
        backup_config = self._fetch_backup_config(BackupFiles.ALL_FILES)
        logger.info('Unselecting all files from backup.')
        directory_tree = DirectoryTree(backup_config.directoryTree)
        directory_tree.unselect_all()
        backup_config.directoryTree = directory_tree.root
        return self._update_backup_config(backup_config)

    def _fetch_backup_config(self, name=None):
        backup_configs = self._edge.api.get('/config/backup/backupPolicy/includeSets')
        if name:
            for backup_config in backup_configs:
                if backup_config.name == name:
                    logger.info('Found backup config: %s', name)
                    return backup_config
            logger.error('Could not find backup config: %s', name)
            raise CTERAException(f'Could not find backup config: {name}')
        return backup_configs

    def _update_backup_config(self, backup_config):
        return self._edge.api.put(f'/config/backup/backupPolicy/includeSets/{backup_config._uuid}', backup_config)  # pylint: disable=W0212



from ..lib.storage import commonfs
from ..exceptions import CTERAException
from .base_command import BaseCommand


class UploadTaskStatus():
    IN_PROGRESS = 0
    COMPLETE = 1
    FAIL = -1


class Firmware(BaseCommand):
    """ Edge Filer Firmware upgrade API """

    def upgrade(self, file_path, reboot=True, wait_for_reboot=True):
        """
        Upgrade the Filer firmware with the provided file

        :param str file_path: Path to the local file to upload
        :param bool,optional reboot: Perform reboot after uploading the new firmware, defaults to True
        :param bool,optional wait_for_reboot: Wait for reboot to complete (if reboot is performed), defaults to True
        """
        upload_task_info = self._upload_firmware(file_path)
        if upload_task_info.rc != 0:
            raise CTERAException(f'Failed to upload firmware: {file_path}')
        self._wait_for_completion(upload_task_info.taskPointer)
        if reboot:
            self._edge.power.reboot(wait=wait_for_reboot)

    def _upload_firmware(self, file_path):
        commonfs.properties(file_path)
        with open(file_path, 'rb') as fd:
            return self._edge.api.form_data(
                'proc/firmware',
                dict(
                    name='upload',
                    firmware=fd
                )
            )

    def _wait_for_completion(self, task_pointer):
        while True:
            task_status = self._edge.api.get(task_pointer)
            is_running = task_status.status == UploadTaskStatus.IN_PROGRESS
            if not is_running:
                if task_status.status == UploadTaskStatus.COMPLETE:
                    return
                raise CTERAException(f'An error occurred during firmware upgrade. Status: {task_status.statusMessage}')



class XMLTypes:
    OBJ = 'obj'
    ATT = 'att'
    VAL = 'val'
    LIST = 'list'
    ID = 'id'
    CLASS = 'class'
    UUID = 'uuid'
    DB = 'db'
    VERSION = 'version'
    FIRMWARE = 'firmware'
    NS = 'xmlns:xsi'
    LOCATION = 'xsi:noNamespaceSchemaLocation'


class DAVTypes:
    HREF = 'href'
    PROPSTAT = 'propstat'
    PROP = 'prop'
    RESPONSE = 'response'
    MULTISTATUS = 'multistatus'
    CREATED_DATE = 'creationdate'
    LAST_MODIFIED = 'getlastmodified'
    CONTENT_TYPE = 'getcontenttype'
    CONTENT_LENGTH = 'getcontentlength'



from .platform import Platform


class Version:  # pylint: disable=unused-private-member

    __instance = None

    @staticmethod
    def instance():
        if Version.__instance is None:
            Version()
        return Version.__instance

    def __init__(self):
        if Version.__instance is not None:
            raise Exception("Version is a singleton class.")

        self.product = 'Chopin'
        self.product_version = '3.0'
        self.system = Platform.instance().os()
        self.machine = Platform.instance().arch()
        self.header = \
            f"{self.product}/{self.product_version} ({self.system}; {self.machine}) Python-urllib/{Platform.instance().python_version()}"
        Version.__instance = self

    def as_header(self):

        return self.header



from ..base_command import BaseCommand
from ....cio.edge import EdgePath
from ....lib.storage import asynfs, commonfs
from . import io


class FileBrowser(BaseCommand):
    """ Edge Filer File Browser APIs """

    async def listdir(self, path):
        """
        List Directory

        :param str path: Path
        """
        return await io.listdir(self._edge, self.normalize(path))

    async def walk(self, path=None):
        """
        Walk Directory Contents

        :param str, defaults to the root directory path: Path to walk
        """
        return io.walk(self._edge, path)

    async def exists(self, path):
        """
        Check if item exists

        :param str path: Path
        """
        return await io.exists(self._edge, self.normalize(path))

    async def handle(self, path):
        """
        Get File Handle.

        :param str path: Path to a file
        """
        handle_function = io.handle(self.normalize(path))
        return await handle_function(self._edge)

    async def handle_many(self, directory, *objects):
        """
        Get a Zip Archive File Handle.

        :param str directory: Path to a folder
        :param args objects: List of files and folders
        """
        handle_many_function = io.handle_many(self.normalize(directory), *objects)
        return await handle_many_function(self._edge)

    async def download(self, path, destination=None):
        """
        Download a file

        :param str path: The file path on the Edge Filer
        :param str,optional destination:
         File destination, if it is a directory, the original filename will be kept, defaults to the default directory
        """
        directory, name = commonfs.determine_directory_and_filename(path, destination=destination)
        handle = await self.handle(path)
        return await asynfs.write(directory, name, handle)

    async def download_many(self, target, objects, destination=None):
        """
        Download selected files and/or directories as a ZIP archive.

        .. warning::
            The provided list of objects is not validated. Only existing files and directories
            will be included in the resulting ZIP file.

        :param str target:
            Path to the cloud folder containing the files and directories to download.
        :param list[str] objects:
            List of file and/or directory names to include in the download.
        :param str destination:
            Optional. Path to the destination file or directory. If a directory is provided,
            the original filename will be preserved. Defaults to the default download directory.
        """
        directory, name = commonfs.determine_directory_and_filename(target, objects, destination=destination, archive=True)
        handle = await self.handle_many(target, *objects)
        return await asynfs.write(directory, name, handle)

    async def upload(self, name, destination, handle):
        """
        Upload from file handle.

        :param str name: File name.
        :param str destination: Path to remote directory.
        :param object handle: Handle.
        """
        upload_function = io.upload(name, self.normalize(destination), handle)
        return await upload_function(self._edge)

    async def upload_file(self, path, destination):
        """
        Upload a file.

        :param str path: Local path
        :param str destination: Remote path
        """
        metadata = commonfs.properties(path)
        with open(path, 'rb') as handle:
            response = await self.upload(metadata['name'], destination, handle)
        return response

    async def mkdir(self, path):
        """
        Create a new directory

        :param str path: Directory path
        """
        return await io.mkdir(self._edge, self.normalize(path))

    async def makedirs(self, path):
        """
        Create a directory recursively

        :param str path: Directory path
        """
        return await io.makedirs(self._edge, self.normalize(path))

    async def copy(self, path, destination=None, overwrite=False):
        """
        Copy a file or a folder

        :param str path: Source file or folder path
        :param str destination: Destination folder path
        :param bool,optional overwrite: Overwrite on conflict, defaults to False
        """
        if destination is None:
            raise ValueError('Copy destination was not specified.')
        return await io.copy(self._edge, self.normalize(path), self.normalize(destination), overwrite)

    async def move(self, path, destination=None, overwrite=False):
        """
        Move a file or a folder

        :param str path: Source file or folder path
        :param str destination: Destination folder path
        :param bool,optional overwrite: Overwrite on conflict, defaults to False
        """
        if destination is None:
            raise ValueError('Move destination was not specified.')
        return await io.move(self._edge, self.normalize(path), self.normalize(destination), overwrite)

    async def delete(self, *paths):
        """
        Delete a file

        :param str path: File path
        """
        return await io.remove(self._edge, *[self.normalize(path) for path in paths])

    @staticmethod
    def normalize(path):
        return EdgePath('/', path)



import logging

from .enum import Mode
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class NTP(BaseCommand):
    """ Edge Filer NTP configuration """

    def get_configuration(self):
        return self._edge.api.get('/config/time')

    @property
    def servers(self):
        return self._edge.api.get('/config/time/NTPServer')

    def enable(self, servers=None):
        """
        Enable NTP

        :param list[str] servers: List of NTP servers address
        """
        logger.info("Enabling time synchronization with ntp servers.")
        self._edge.api.put('/config/time/NTPMode', Mode.Enabled)
        logger.info("Time synchronization enabled.")

        if servers:
            logger.info("Updating time servers. %s", {'servers': servers})
            self._edge.api.put('/config/time/NTPServer', servers)
            logger.info("Time servers updated. %s", {'servers': servers})

    def disable(self):
        """ Disable NTP """
        logger.info("Disabling time synchronization with ntp servers.")
        self._edge.api.put('/config/time/NTPMode', Mode.Disabled)
        logger.info("Time synchronization disabled.")



# pylint: disable=line-too-long
import logging

from ..lib import QueryLogsResponse
from . import query
from . import enum
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class Logs(BaseCommand):
    """
    Edge Filer Logs APIs

    :ivar list[str] default_include: Default log fields - 'severity', 'time', 'msg', 'more'
    """

    default_include = ['severity', 'time', 'msg', 'more']

    def settings(self, retention, min_severity=None):
        """
        Configure log settings

        :param int retention: Log retention period in days
        :param cterasdk.edge.enum.Severity,optional min_severity: Minimal log severity
        """
        log_config = self._edge.api.get('/config/logging/general')
        log_config.LogKeepPeriod = retention
        if min_severity:
            log_config.minSeverity = min_severity
        logger.info('Updating log settings. %s', {'retention': retention, 'min_severity': log_config.minSeverity})
        self._edge.api.put('/config/logging/general', log_config)
        logger.info('Log settings updated. %s', {'retention': retention, 'min_severity': log_config.minSeverity})

    def logs(self, topic, include=None, minSeverity=enum.Severity.INFO):
        """
        Fetch Edge Filer logs

        :param str topic: Log Topic to fetch
        :param list[str],optional include: List of fields to include in the response, defailts to Logs.default_include
        :param cterasdk.edge.enum.Severity,optional minSeverity: Minimal log severity to fetch, defaults to cterasdk.edge.enum.Severity.INFO

        :return: Log entries
        :rtype: cterasdk.lib.iterator.QueryIterator
        """
        param = query.QueryParamBuilder().include(
            include or Logs.default_include).put('topic', topic).put('minSeverity', minSeverity).build()
        return query.iterator(self._edge, '/config/logging/general', param, 'pagedQuery', QueryLogsResponse)



from .base import CTERAException


class NotificationsError(CTERAException):
    """
    Notifications error

    :ivar list[int] cloudfolders: List of cloudfolders
    :ivar str cursor: Cursor
    """

    def __init__(self, cloudfolders, cursor):
        super().__init__('An error occurred while trying to retrieve notifications.')
        self.cloudfolders = cloudfolders
        self.cursor = cursor



import logging

from ..common import union, parse_base_object_ref, ApplicationBackupSet, PolicyRuleConverter, Object
from ..exceptions import CTERAException, ObjectNotFoundException
from .base_command import BaseCommand
from . import query
from .enum import Platform


logger = logging.getLogger('cterasdk.core')


class Templates(BaseCommand):
    """
    Portal Configuration Template APIs
    """

    default = ['name']

    def __init__(self, portal):
        super().__init__(portal)
        self.auto_assign = TemplateAutoAssignPolicy(self._core)

    def _get_entire_object(self, name):
        ref = f'/deviceTemplates/{name}'
        try:
            return self._core.api.get(ref)
        except CTERAException as error:
            raise CTERAException(f'Template not found: {ref}') from error

    def get(self, name, include=None):
        """
        Get a Configuration Template

        :param str name: Name of the template
        :param list[str] include: List of fields to retrieve, defaults to ``['name']``
        """
        include = union(include or [], Templates.default)
        include = ['/' + attr for attr in include]
        template = self._core.api.get_multi('/deviceTemplates/' + name, include)
        if template.name is None:
            raise ObjectNotFoundException(f'/deviceTemplates/{name}')
        return template

    def add(self, name, description=None, include_sets=None, exclude_sets=None,  # pylint: disable=too-many-arguments
            apps=None, backup_schedule=None, versions=None, update_settings=None,
            scripts=None, cli_commands=None, consent_page=None):
        """
        Add a Configuration Template

        :param str name: Name of the template
        :param str,optional description: Template description
        :param list[cterasdk.common.types.FilterBackupSet],optional include_sets: List of backup sets to include
        :param list[cterasdk.common.types.FilterBackupSet],optional exclude_sets: List of backup sets to exclude
        :param list[cterasdk.common.enum.Application],optional apps: List of applications to back up
        :param cterasdk.common.types.TaskSchedule,optional backup_schedule: Backup schedule
        :param list[cterasdk.core.types.PlatformVersion],optional versions: List of platforms and their associated versions.
         Pass `None` to inherit the default settings from the Global Administration Portal
        :param cterasdk.common.types.SoftwareUpdatesTopic,optional update_settings: Software update settings
        :param list[cterasdk.core.types.TemplateScript],optional scripts: Scripts to execute after logon, before or after backup
        :param list[str],optional cli_commands: Template CLI commands to execute
        :param cterasdk.common.types.ConsentPage consent_page: Consent page to show to end-user
        """
        param = Object()
        param._classname = 'DeviceTemplate'  # pylint: disable=protected-access
        param.name = name
        param.description = description

        self._configure_firmware_settings(param, versions)

        param.deviceSettings = Object()
        param.deviceSettings._classname = 'DeviceTemplateSettings'  # pylint: disable=protected-access

        Templates._configure_backup_settings(param, include_sets, exclude_sets, backup_schedule, apps)
        Templates._add_scripts(param, scripts)
        Templates._add_cli_commands(param, cli_commands)
        Templates._configure_software_update_schedule(param, update_settings)
        Templates._configure_consent_page(param, consent_page)

        logger.info('Adding template. %s', {'name': name})
        response = self._core.api.add('/deviceTemplates', param)
        logger.info('Template added. %s', {'name': name})
        return response

    def _configure_firmware_settings(self, param, versions):
        param.firmwaresSettings = Object()
        param.firmwaresSettings._classname = 'FirmwaresSettings'  # pylint: disable=protected-access

        if versions:
            param.firmwaresSettings.useGlobal = False
            param.firmwaresSettings.firmwares = self._convert_to_template_firmwares(versions)
        else:
            param.firmwaresSettings.useGlobal = True
            param.firmwaresSettings.firmwares = None

    @staticmethod
    def _configure_backup_settings(param, include_sets, exclude_sets, backup_schedule, apps):
        if include_sets or exclude_sets or backup_schedule or apps:
            param.deviceSettings.backup = Object()
            param.deviceSettings.backup._classname = 'BackupTemplate'  # pylint: disable=protected-access
            param.deviceSettings.backup.backupPolicy = Object()
            param.deviceSettings.backup.backupPolicy._classname = 'BackupPolicyTemplate'  # pylint: disable=protected-access
            if include_sets:
                param.deviceSettings.backup.backupPolicy.includeSets = include_sets
            if exclude_sets:
                param.deviceSettings.backup.backupPolicy.excludeSets = exclude_sets
            if backup_schedule:
                param.deviceSettings.backup.scheduleTopic = Object()
                param.deviceSettings.backup.scheduleTopic._classname = 'ScheduleTopic'  # pylint: disable=protected-access
                param.deviceSettings.backup.scheduleTopic.overrideTemplate = True
                param.deviceSettings.backup.scheduleTopic.schedule = backup_schedule
            if apps:
                param.deviceSettings.backup.applicationsTopic = Object()
                param.deviceSettings.backup.applicationsTopic._classname = 'ApplicationsTopic'  # pylint: disable=protected-access
                param.deviceSettings.backup.applicationsTopic.overrideTemplate = True
                param.deviceSettings.backup.applicationsTopic.includeApps = ApplicationBackupSet(apps)

    @staticmethod
    def _configure_software_update_schedule(param, update_settings):
        if update_settings:
            param.deviceSettings.softwareUpdates = Object()
            param.deviceSettings.softwareUpdates._classname = 'SoftwareUpdatesTopic'  # pylint: disable=protected-access
            param.deviceSettings.softwareUpdates.overrideTemplate = False
            param.deviceSettings.softwareUpdates.softwareUpdates = update_settings

    @staticmethod
    def _configure_consent_page(param, consent_page):
        if consent_page:
            param.deviceSettings.consentPage = Object()
            param.deviceSettings.consentPage._classname = 'ConsentPageTopic'  # pylint: disable=protected-access
            param.deviceSettings.consentPage.enabled = True
            param.deviceSettings.consentPage.consentPageHeader = consent_page.header
            param.deviceSettings.consentPage.consentPageBody = consent_page.body

    @staticmethod
    def _add_scripts(param, scripts):
        if scripts:
            param.deviceSettings.scripts = Object()
            param.deviceSettings.scripts._classname = 'ScriptTemplates'  # pylint: disable=protected-access
            for script in scripts:
                server_object = script.to_server_object()
                if script.platform == Platform.Windows:
                    param.deviceSettings.scripts.windowsScripts = server_object
                if script.platform == Platform.Linux:
                    param.deviceSettings.scripts.linuxScripts = server_object
                if script.platform == Platform.OSX:
                    param.deviceSettings.scripts.macScripts = server_object

    @staticmethod
    def _add_cli_commands(param, cli_commands):
        if cli_commands:
            param.deviceSettings.cliCommands = Object()
            param.deviceSettings.cliCommands._classname = 'CliCommandTemplate'  # pylint: disable=protected-access
            param.deviceSettings.cliCommands.cliCommands = cli_commands

    def _convert_to_template_firmwares(self, versions):
        firmwares = {image.name: parse_base_object_ref(image.baseObjectRef) for image in self._core.firmwares.list_images()}

        template_firmwares = []
        for platform, version in versions:
            base_object_ref = firmwares.get(f'{platform}-{version}')
            if base_object_ref is None:
                raise CTERAException(f"No firmware found for platform '{platform}' and version '{version}'.")
            template_firmwares.append(Templates._create_template_firmware(platform, str(base_object_ref)))

        return template_firmwares

    @staticmethod
    def _create_template_firmware(platform, base_object_ref):
        param = Object()
        param._classname = 'TemplateFirmware'  # pylint: disable=protected-access
        param.type = platform
        param.firmware = base_object_ref
        return param

    def by_name(self, names, include=None):
        """
        Get Templates by their names

        :param list[str],optional names: List of names of templates
        :param list[str],optional include: List of fields to retrieve, defaults to ['name']

        :return: Iterator for all matching Templates
        :rtype: cterasdk.lib.iterator.Iterator
        """
        filters = [query.FilterBuilder('name').eq(name) for name in names]
        return self.list_templates(include, filters)

    def list_templates(self, include=None, filters=None):
        """
        List Configuration Templates.\n
        To retrieve templates, you must first browse the tenant, using: `GlobalAdmin.portals.browse()`

        :param list[str],optional include: List of fields to retrieve, defaults to ``['name']``
        :param list[],optional filters: List of additional filters, defaults to None
        """
        include = union(include or [], Templates.default)
        builder = query.QueryParamBuilder().include(include)
        if filters:
            for query_filter in filters:
                builder.addFilter(query_filter)
            builder.orFilter((len(filters) > 1))
        param = builder.build()
        return query.iterator(self._core, '/deviceTemplates', param)

    def delete(self, name):
        """
        Delete a Configuration Template

        :param str name: Name of the template
        """
        logger.info('Deleting template. %s', {'name': name})
        response = self._core.api.delete(f'/deviceTemplates/{name}')
        logger.info('Template deleted. %s', {'name': name})
        return response

    def set_default(self, name, wait=False):
        """
        Set a Configuration Template as the default template

        :param str name: Name of the template
        :param bool,optional wait: Wait for all changes to apply, defaults to `False`
        """
        logger.info('Setting default template. %s', {'name': name})
        response = self._core.api.execute(f'/deviceTemplates/{name}', 'setAsDefault')
        self.auto_assign.apply_changes(wait=wait)
        logger.info('Set default template. %s', {'name': name})
        return response

    def remove_default(self, name, wait=False):
        """
        Set a Configuration Template not to be the default template

        :param str name: Name of the template
        :param bool,optional wait: Wait for all changes to apply, defaults to `False`
        """
        template = self.get(name, include=['isDefault'])
        if template.isDefault:
            logger.info('Removing default template. %s', {'name': name})
            response = self._core.api.execute('', 'removeDefaultDeviceTemplate')
            logger.info('Removed default template. %s', {'name': name})
            self.auto_assign.apply_changes(wait=wait)
            return response
        logger.info('Template not set as default. %s', {'name': name})
        return None


class TemplateAutoAssignPolicy(BaseCommand):

    def get_policy(self):
        """
        Get templates auto assignment policy
        """
        return self._core.api.execute('', 'getAutoAssignmentRules')

    def set_policy(self, rules, apply_default=None, default=None, apply_changes=True):
        """
        Set templates auto assignment policy

        :param list[cterasdk.common.types.PolicyRule] rules: List of policy rules
        :param bool,optional apply_default: If no match found, apply default template. If not passed, the current config will be kept
        :param str,optional default: Name of a template to assign if no match found. Ignored unless the ``apply_default`` is set to ``True``
        :param bool,optional apply_changes: Apply changes upon update, defaults to ``True``
        """
        templates = {rule.assignment for rule in rules}
        if default:
            templates.add(default)
        templates = list(templates)
        portal_templates = {template.name: template for template in self._core.templates.by_name(templates, ['baseObjectRef'])}

        not_found = [template for template in templates if template not in portal_templates.keys()]
        if not_found:
            logger.error('Could not find one or more templates: %s', not_found)
            raise CTERAException(f'Could not find one or more templates: {not_found}.')

        policy = self.get_policy()

        if apply_default is False:
            policy.defaultTemplate = None
        elif apply_default is True and default:
            policy.defaultTemplate = portal_templates.get(default).baseObjectRef

        policy_rules = [PolicyRuleConverter.convert(rule, 'DeviceTemplateAutoAssignmentRule', 'template',
                        portal_templates.get(rule.assignment).baseObjectRef) for rule in rules]
        policy.deviceTemplatesAutoAssignmentRules = policy_rules

        response = self._core.api.execute('', 'setAutoAssignmentRules', policy)
        logger.info('Set templates auto assignment rules.')

        if apply_changes:
            self.apply_changes(True)

        return response

    def apply_changes(self, wait=False):
        """
        Apply provisioning changes.\n

        :param bool,optional wait: Wait for all changes to apply, defaults to `False`
        """
        task = self._core.api.execute('', 'applyAutoAssignmentRules')
        if wait:
            task = self._core.tasks.wait(task)
        return task



import logging

from .base_command import BaseCommand
from ..exceptions import CTERAException


logger = logging.getLogger('cterasdk.core')


class Login(BaseCommand):
    """
    Portal Login APIs
    """

    def login(self, username, password):
        """
        Log into the portal

        :param str username: User name to log in
        :param str password: User password
        """
        host = self._core.host()
        try:
            self._core.api.form_data('/login', {'j_username': username, 'j_password': password})
            logger.info("User logged in. %s", {'host': host, 'user': username})
        except CTERAException:
            logger.error('Login failed. %s', {'host': host, 'user': username})
            raise

    def sso(self, ctera_ticket):
        """
        Login using a Portal ticket.

        :param str ticket: SSO Ticket.
        """
        logger.info('Logging in using a Portal ticket.')
        self._core.ctera.form_data('/sso', {'ctera_ticket': ctera_ticket})

    def logout(self):
        """
        Log out of the portal
        """
        username = self._core.session().account.name
        self._core.api.form_data('/logout', {})
        logger.info("User logged out. %s", {'host': self._core.host(), 'user': username})



import logging

from ..common import Object
from ..exceptions import CTERAException
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class Users(BaseCommand):
    """ Edge Filer Users configuration APIs """

    def get(self, name=None):
        """
        Get User. If a user name was not passed as an argument, a list of all local users will be retrieved

        :param str,optional name: Name of the user
        """
        return self._edge.api.get('/config/auth/users' + ('' if name is None else ('/' + name)))

    def add_first_user(self, username, password, email=''):
        """
        Add the first user of the Edge Filer and login

        :param str username: User name for the new user
        :param str password: Password for the new user
        :param str,optional email: E-mail address of the new user, defaults to an empty string
        """
        if not self._edge.initialized:
            user = Object()
            user.username = username
            user.password = password
            user.email = email
            self._edge.api.post('/nosession/createfirstuser', user)
            logger.info('User created: %s', username)
        else:
            logger.info('User already exists.')
        self._edge.login(username, password)

    def add(self, username, password, full_name=None, email=None, uid=None):
        """
        Add a user of the Gateway

        :param str username: User name for the new user
        :param str password: Password for the new user
        :param str,optional full_name: The full name of the new user, defaults to None
        :param str,optional email: E-mail address of the new user, defaults to None
        :param str,optional uid: The uid of the new user, defaults to None
        """
        user = Object()
        user.username = username
        user.password = password
        user.fullName = full_name
        user.email = email
        user.uid = uid
        try:
            response = self._edge.api.add('/config/auth/users', user)
            logger.info("User created: %s", user.username)
            return response
        except CTERAException as error:
            logger.error("User creation failed: %s", username)
            raise CTERAException(f'User creation failed: {username}') from error

    def modify(self, username, password=None, full_name=None, email=None, uid=None):
        """
        Modify an existing user of the Gateway

        :param str username: User name to modify
        :param str,optional password: New password, defaults to None
        :param str,optional full_name: The full name of the user, defaults to None
        :param str,optional email: E-mail address of the user, defaults to None
        :param str,optional uid: The uid of the user, defaults to None
        """
        ref = f'/config/auth/users/{username}'
        try:
            user = self._edge.api.get(ref)
        except CTERAException as error:
            raise CTERAException(f'User not found: {ref}') from error

        if password:
            user.password = password
        if full_name:
            user.fullName = full_name
        if email:
            user.email = email
        if uid:
            user.uid = uid
        try:
            response = self._edge.api.put(ref, user)
            logger.info("User modified: %s", ref)
            return response
        except CTERAException as error:
            logger.error('User modification failed: %s', ref)
            raise CTERAException(f'User modification failed: {ref}') from error

    def delete(self, username):
        """
        Delete an existing user

        :param str username: User name of the user to delete
        """
        ref = f'/config/auth/users/{username}'
        try:
            response = self._edge.api.delete(ref)
            logger.info("User deleted: %s", ref)
            return response
        except Exception as error:
            logger.error("User deletion failed: %s", ref)
            raise CTERAException(f'User deletion failed: {ref}') from error



import logging
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.core')


class Settings(BaseCommand):
    """
    CTERA Portal Settings APIs

    :ivar cterasdk.asynchronous.core.settings.GlobalSettings global_settings: Object holding the Portal Global Settings APIs
    """

    def __init__(self, core):
        super().__init__(core)
        self.global_settings = GlobalSettings(core)


class GlobalSettings(BaseCommand):
    """
    Global Settings APIs
    """

    @property
    async def dns_suffix(self):
        return await self._core.v1.api.get('/settings/dnsSuffix')



import logging

from .enum import Mode
from .base_command import BaseCommand


logger = logging.getLogger('cterasdk.edge')


class AFP(BaseCommand):
    """ Edge Filer AFP APIs """

    def is_disabled(self):
        """ Check if the AFP server is disabled """
        return self._edge.api.get('/config/fileservices/afp/mode') == Mode.Disabled

    def disable(self):
        """
        Disable AFP
        """
        logger.info('Disabling AFP server.')
        self._edge.api.put('/config/fileservices/afp/mode', Mode.Disabled)
        logger.info('AFP server disabled.')



from abc import ABC
from collections import namedtuple
from ..common import DateTimeUtils, StringCriteriaBuilder, PredefinedListCriteriaBuilder, CustomListCriteriaBuilder, Object
from ..lib.storage import commonfs

from .enum import PortalAccountType, CollaboratorType, FileAccessMode, PlanCriteria, TemplateCriteria, \
                  BucketType, LocationType, Platform, RetentionMode, Duration, ExtendedAttributes


CloudFSFolderFindingHelper = namedtuple('CloudFSFolderFindingHelper', ('name', 'owner'))
CloudFSFolderFindingHelper.__doc__ = 'Tuple holding the name and owner couple to search for folders'
CloudFSFolderFindingHelper.name.__doc__ = 'The name of the CloudFS folder'
CloudFSFolderFindingHelper.owner.__doc__ = 'The name of the owner of the CloudFS folder'


PlatformVersion = namedtuple('PlatformVersion', ('name', 'version'))
PlatformVersion.__doc__ = 'Tuple holding the platform name and version'
PlatformVersion.name.__doc__ = 'The name of the platform'
PlatformVersion.version.__doc__ = 'The version identifier'


AccessControlEntry = namedtuple('AccessControlEntry', ('account', 'role'))
AccessControlEntry.__doc__ = 'Tuple holding a Portal account and its respective permission'
AccessControlEntry.account.__doc__ = 'The Portal group or user account'
AccessControlEntry.role.__doc__ = 'The group or user role'


class PortalAccount(ABC):
    """
    Base Class for Portal Account

    :ivar str name: The user name
    :ivar str directory: The fully-qualified name of the user directory, defaults to None
    """
    def __init__(self, name, directory=None):
        """
        :param str name: The name of the Portal user
        :param str directory: The the fully qualified domain name, defaults to None
        """
        self.name = name
        self.directory = directory

    @property
    def is_local(self):
        """
        Is the account local

        :return bool: True if the account if local, otherwise False
        """
        return not self.directory

    @property
    def account_type(self):
        """
        The Portal Account Type

        :return cterasdk.core.enum.PortalAccountType: The Portal Account Type
        """
        raise NotImplementedError('Implementing class much implement the account_type property')

    @staticmethod
    def from_collaborator(collaborator):
        if collaborator.type == CollaboratorType.LU:
            return UserAccount(collaborator.name)
        if collaborator.type == CollaboratorType.DU:
            return UserAccount(collaborator.name, collaborator.domain)
        if collaborator.type == CollaboratorType.LG:
            return GroupAccount(collaborator.name)
        if collaborator.type == CollaboratorType.DG:
            return GroupAccount(collaborator.name, collaborator.domain)
        return None

    def __eq__(self, other):
        if type(self) is type(other):
            if self.account_type == other.account_type and self.name == other.name and self.directory == other.directory:
                return True
        return False

    def __str__(self):
        return (self.directory if self.directory else '') + '\\' + self.name


class UserAccount(PortalAccount):
    @property
    def account_type(self):
        return PortalAccountType.User


class GroupAccount(PortalAccount):
    @property
    def account_type(self):
        return PortalAccountType.Group


class ShareRecipient:
    """
    Class Representing a Collboration Share Recipient
    """
    def __init__(self, account, account_type, two_factor=False):
        self.account = account
        self.type = account_type
        self.two_factor = two_factor
        self.access = None
        self.expiration_date = None

    @staticmethod
    def external(email, two_factor=False):
        """
        Share with an external user

        :param str email: The email address of the recipient
        :param bool two_factor: Require two factor authentication over e-mail
        """
        return ShareRecipient(email, CollaboratorType.EXT, two_factor)

    @staticmethod
    def local_user(user_account):
        """
        Share with a local user

        :param UserAccount user_account: A local user account
        """
        return ShareRecipient(user_account, CollaboratorType.LU)

    @staticmethod
    def domain_user(user_account):
        """
        Share with a domain user

        :param UserAccount user_account: A domain user account
        """
        return ShareRecipient(user_account, CollaboratorType.DU)

    @staticmethod
    def local_group(group_account):
        """
        Share with a local group

        :param GroupAccount group_account: A local group account
        """
        return ShareRecipient(group_account, CollaboratorType.LG)

    @staticmethod
    def domain_group(group_account):
        """
        Share with a domain group

        :param GroupAccount group_account: A domain group account
        """
        return ShareRecipient(group_account, CollaboratorType.DG)

    def read_write(self):
        """
        Grant read write access
        """
        self.access = FileAccessMode.RW
        return self

    def read_only(self):
        """
        Grant read only access
        """
        self.access = FileAccessMode.RO
        return self

    def preview_only(self):
        """
        Grant preview only access
        """
        self.access = FileAccessMode.PO
        return self

    def no_access(self):
        """
        Deny access
        """
        self.access = FileAccessMode.NA
        return self

    def expire_in(self, days):
        """
        Set share to expire after (days)

        :param int days: The number of days the share will remain accessible
        """
        expiration_date = DateTimeUtils.get_expiration_date(days).strftime('%Y-%m-%d')
        return self.expire_on(expiration_date)

    def expire_on(self, expiration_date):
        """
        Set the share expiration date

        :param str expire_on: The expiration date (%Y-%m-%d)
        """
        self.expiration_date = expiration_date
        return self

    def __str__(self):
        if self.type == CollaboratorType.EXT:
            return self.account
        return str(self.account)


class PlanCriteriaBuilder:

    Type = 'PlanCriteria'

    @staticmethod
    def username():
        return StringCriteriaBuilder(PlanCriteriaBuilder.Type, PlanCriteria.Username)

    @staticmethod
    def user_groups():
        return CustomListCriteriaBuilder(PlanCriteriaBuilder.Type, PlanCriteria.Groups)

    @staticmethod
    def role():
        return PredefinedListCriteriaBuilder(PlanCriteriaBuilder.Type, PlanCriteria.Role)

    @staticmethod
    def first_name():
        return StringCriteriaBuilder(PlanCriteriaBuilder.Type, PlanCriteria.First)

    @staticmethod
    def last_name():
        return StringCriteriaBuilder(PlanCriteriaBuilder.Type, PlanCriteria.Last)

    @staticmethod
    def company():
        return StringCriteriaBuilder(PlanCriteriaBuilder.Type, PlanCriteria.Company)

    @staticmethod
    def billing_id():
        return StringCriteriaBuilder(PlanCriteriaBuilder.Type, PlanCriteria.BillingId)

    @staticmethod
    def comment():
        return StringCriteriaBuilder(PlanCriteriaBuilder.Type, PlanCriteria.Comment)


class TemplateCriteriaBuilder:

    Type = 'DeviceCriteria'

    @staticmethod
    def type():
        return PredefinedListCriteriaBuilder(TemplateCriteriaBuilder.Type, TemplateCriteria.Type)

    @staticmethod
    def os():
        return StringCriteriaBuilder(TemplateCriteriaBuilder.Type, TemplateCriteria.OperatingSystem)

    @staticmethod
    def version():
        return StringCriteriaBuilder(TemplateCriteriaBuilder.Type, TemplateCriteria.Version)

    @staticmethod
    def hostname():
        return StringCriteriaBuilder(TemplateCriteriaBuilder.Type, TemplateCriteria.Hostname)

    @staticmethod
    def name():
        return StringCriteriaBuilder(TemplateCriteriaBuilder.Type, TemplateCriteria.Name)

    @staticmethod
    def owner():
        return StringCriteriaBuilder(TemplateCriteriaBuilder.Type, TemplateCriteria.Owner)

    @staticmethod
    def plan():
        return StringCriteriaBuilder(TemplateCriteriaBuilder.Type, TemplateCriteria.Plan)

    @staticmethod
    def groups():
        return StringCriteriaBuilder(TemplateCriteriaBuilder.Type, TemplateCriteria.Groups)


class Bucket:

    def __init__(self, bucket, driver):
        self.bucket = bucket
        self.driver = driver

    def to_server_object(self):
        param = Object()
        param.bucket = self.bucket
        param.storage = self.driver
        return param


class HTTPBucket(Bucket):

    def __init__(self, bucket, driver, access_key, secret_key, endpoint, https, direct=False, verify_ssl=True):
        super().__init__(bucket, driver)
        self.access_key = access_key
        self.secret_key = secret_key
        self.endpoint = endpoint
        self.https = https
        self.direct = direct
        self.verify_ssl = verify_ssl

    @property
    def trust_all_certificates(self):
        return not self.verify_ssl


class AzureBlob(HTTPBucket):

    def __init__(self, bucket, access_key, secret_key, endpoint='core.windows.net', https=True, direct=True):
        super().__init__(bucket, BucketType.Azure, access_key, secret_key, endpoint, https, direct)

    def to_server_object(self):
        param = super().to_server_object()
        param._classname = LocationType.Azure  # pylint: disable=protected-access
        param.endPoint = self.endpoint
        param.accountName = self.access_key
        param.secretAccess = self.secret_key
        param.useHttps = self.https
        param.directUpload = self.direct
        return param


class S3Compatible(HTTPBucket):

    def __init__(self, bucket, driver, access_key, secret_key,
                 endpoint, https, direct, verify_ssl):
        super().__init__(bucket, driver, access_key, secret_key, endpoint, https, direct, verify_ssl)

    def to_server_object(self):
        param = super().to_server_object()
        param._classname = LocationType.S3Compatible  # pylint: disable=protected-access
        param.endPoint = self.endpoint
        param.awsAccessKey = self.access_key
        param.awsSecretKey = self.secret_key
        param.useHttps = self.https
        param.directUpload = self.direct
        param.trustAllCertificates = self.trust_all_certificates
        return param


class Scality(S3Compatible):

    def __init__(self, bucket, access_key, secret_key,
                 endpoint, https=False, direct=False, verify_ssl=True):
        super().__init__(bucket, BucketType.Scality, access_key, secret_key, endpoint, https, direct, verify_ssl)


class ICOS(S3Compatible):

    def __init__(self, bucket, access_key, secret_key,
                 endpoint, https=False, direct=False, verify_ssl=True):
        super().__init__(bucket, BucketType.ICOS, access_key, secret_key, endpoint, https, direct, verify_ssl)


class Nutanix(S3Compatible):

    def __init__(self, bucket, access_key, secret_key,
                 endpoint, https=False, direct=False, verify_ssl=True):
        super().__init__(bucket, BucketType.Nutanix, access_key, secret_key, endpoint, https, direct, verify_ssl)


class Wasabi(S3Compatible):

    def __init__(self, bucket, access_key, secret_key,
                 endpoint, https=False, direct=False, verify_ssl=True):
        super().__init__(bucket, BucketType.Wasabi, access_key, secret_key, endpoint, https, direct, verify_ssl)


class Google(S3Compatible):

    def __init__(self, bucket, access_key, secret_key,
                 endpoint, https=False, direct=False, verify_ssl=True):
        super().__init__(bucket, BucketType.Google, access_key, secret_key, endpoint, https, direct, verify_ssl)


class GenericS3(S3Compatible):

    def __init__(self, bucket, access_key, secret_key,
                 endpoint, https=False, direct=False, verify_ssl=True):
        super().__init__(bucket, BucketType.GenericS3, access_key, secret_key, endpoint, https, direct, verify_ssl)


class NetAppStorageGRID(S3Compatible):

    def __init__(self, bucket, access_key, secret_key,
                 endpoint, https=False, direct=False, tags=False, verify_ssl=True):
        super().__init__(bucket, BucketType.NetAppStorageGRID, access_key, secret_key, endpoint, https, direct, verify_ssl)
        self.tagBlocks = tags

    def to_server_object(self):
        param = super().to_server_object()
        param._classname = LocationType.NetAppStorageGRID  # pylint: disable=protected-access
        return param


class AmazonS3(HTTPBucket):

    def __init__(self, bucket, access_key=None, secret_key=None, endpoint='s3.amazonaws.com', https=True, direct=True, verify_ssl=True):
        super().__init__(bucket, BucketType.AWS, access_key, secret_key, endpoint, https, direct, verify_ssl)

    def to_server_object(self):
        param = super().to_server_object()
        param._classname = LocationType.S3  # pylint: disable=protected-access
        param.s3Endpoint = self.endpoint
        param.awsAccessKey = self.access_key
        param.awsSecretKey = self.secret_key
        param.httpsOnly = self.https
        param.directUpload = self.direct
        param.trustAllCertificates = self.trust_all_certificates
        return param


class DomainControllers:

    def __init__(self, primary=None, secondary=None):
        self._primary = primary
        self._secondary = secondary

    @property
    def primary(self):
        return self._primary

    @property
    def secondary(self):
        return self._secondary


class AccessControlRule(Object):

    def __init__(self, group, role):
        super().__init__()
        self._classname = 'AccessControlRule'
        self.group = group
        self.role = role


class TemplateScript:

    def __init__(self, platform):
        self._platform = platform
        self._after_logon = None
        self._before_backup = None
        self._after_backup = None

    @property
    def platform(self):
        return self._platform

    @staticmethod
    def windows():
        """
        Configure Windows Scripts
        """
        return TemplateScript(Platform.Windows)

    @staticmethod
    def linux():
        """
        Configure Windows Scripts
        """
        return TemplateScript(Platform.Linux)

    @staticmethod
    def mac():
        """
        Configure Windows Scripts
        """
        return TemplateScript(Platform.OSX)

    def after_logon(self, after_logon):
        """
        Set the post logon script

        :param str after_logon: A string or path to the script file
        """
        self._after_logon = TemplateScript._get_contents(after_logon)
        return self

    def before_backup(self, before_backup):
        """
        Set the pre backup script

        :param str before_backup: A string or path to the script file
        """
        self._before_backup = TemplateScript._get_contents(before_backup)
        return self

    def after_backup(self, after_backup):
        """
        Set the post backup script

        :param str after_backup: A string or path to the script file
        """
        self._after_backup = TemplateScript._get_contents(after_backup)
        return self

    @staticmethod
    def _get_contents(shell_script):
        if commonfs.exists(shell_script):
            with open(shell_script, 'r', encoding='utf-8') as f:
                data = f.read()
            return data
        return shell_script

    def to_server_object(self):
        param = Object()
        param._classname = 'OsScriptTemplates'  # pylint: disable=protected-access
        if self._before_backup is not None:
            param.beforeBackup = self._before_backup
        if self._after_backup is not None:
            param.afterBackup = self._after_backup
        if self._after_logon is not None:
            param.afterFirtSignIn = self._after_logon
        return param


class AlertBuilder:

    def __init__(self, name):
        self.param = Object()
        self.param._classname = 'AlertRule'
        self.param.id = name

    @staticmethod
    def name(name):
        """
        Create an Alert Builder

        :param str name: Alert name
        :returns: Alert Builder
        :rtype: cterasdk.core.types.AlertBuilder
        """
        return AlertBuilder(name)

    def description(self, description):
        """
        Set alert description

        :param str description: Alert description
        :returns: Alert Builder
        :rtype: cterasdk.core.types.AlertBuilder
        """
        self.param.description = description
        return self

    def log(self, log):
        """
        Set alert log class name

        :param str log: Log class name
        :returns: Alert Builder
        :rtype: cterasdk.core.types.AlertBuilder
        """
        self.param.logName = log
        return self

    def topic(self, topic):
        """
        Set alert log topic

        :param cterasdk.core.enum.LogTopic topic: Log topic
        :returns: Alert Builder
        :rtype: cterasdk.core.types.AlertBuilder
        """
        self.param.topic = topic
        return self

    def min_severity(self, min_severity):
        """
        Set alert log minimum severity

        :param cterasdk.core.enum.Severity min_severity: Minimum severity
        :returns: Alert Builder
        :rtype: cterasdk.core.types.AlertBuilder
        """
        self.param.minSeverity = min_severity
        return self

    def origin_type(self, origin_type):
        """
        Set alert origin type

        :param cterasdk.core.enum.OriginType origin_type: Log origin type
        :returns: Alert Builder
        :rtype: cterasdk.core.types.AlertBuilder
        """
        self.param.originType = origin_type
        return self

    def content(self, content):
        """
        Set alert log message content

        :param str content: Log content
        :returns: Alert Builder
        :rtype: cterasdk.core.types.AlertBuilder
        """
        self.param.messageContent = content
        return self

    def build(self):
        """
        Build the alert

        :returns: Alert object
        :rtype: cterasdk.common.object.Object
        """
        return self.param


class Task(Object):

    def __init__(self, task_id, name):
        super().__init__()
        self.id = task_id
        self.name = name


class ScheduledTask(Task):

    def __init__(self, task_id, name, start_time):
        super().__init__(task_id, name)
        self.start_time = start_time

    @staticmethod
    def from_server_object(server_object):
        return ScheduledTask(
            server_object.id,
            server_object.name,
            server_object.startTime,
        )


class BackgroundTask(Task):

    def __init__(self, task_id, name, start_time, end_time, elapsed_time, status, message, ref):
        super().__init__(task_id, name)
        self.start_time = start_time
        self.end_time = end_time
        self.elapsed_time = elapsed_time
        self.status = status
        self.message = message
        self.ref = ref

    @staticmethod
    def from_server_object(server_object, ref):
        return BackgroundTask(
            server_object.id,
            server_object.name,
            server_object.startTime,
            server_object.endTime,
            server_object.elapsedTime,
            server_object.status,
            server_object.progstring,
            ref
        )


class ComplianceSettingsBuilder:

    def __init__(self, enabled, mode, retain_for):
        self.settings = Object()
        self.settings._classname = 'WormSettings'  # pylint: disable=protected-access
        self.settings.worm = enabled
        self.settings.retentionMode = mode
        self.settings.retentionPeriod = retain_for
        self.settings.gracePeriod = None

    @staticmethod
    def default():
        return ComplianceSettingsBuilder(False, RetentionMode.Delete, None)

    @staticmethod
    def none(amount, duration):
        return ComplianceSettingsBuilder(True, RetentionMode.Delete, ComplianceSettingsBuilder._get_retention_period(amount, duration))

    @staticmethod
    def enterprise(amount, duration):
        return ComplianceSettingsBuilder(True, RetentionMode.Enterprise, ComplianceSettingsBuilder._get_retention_period(amount, duration))

    @staticmethod
    def compliance(amount, duration):
        return ComplianceSettingsBuilder(True, RetentionMode.Compliance, ComplianceSettingsBuilder._get_retention_period(amount, duration))

    @staticmethod
    def _get_retention_period(amount, duration):
        retain_for = Object()
        retain_for._classname = 'WormPeriod'  # pylint: disable=protected-access
        retain_for.amount = amount
        retain_for.type = duration
        return retain_for

    def grace_period(self, amount=30, duration=Duration.Minutes):
        self.settings.gracePeriod = self._get_retention_period(amount, duration)
        return self

    def build(self):
        if self.settings.gracePeriod is None:
            self.grace_period()
        return self.settings


class ExtendedAttribute(Object):

    def __init__(self, name, supported):
        super().__init__()
        self._classname = 'ExtendedAttributesInfo'  # pylint: disable=protected-access
        self.name = name
        self.supported = supported


class ExtendedAttributesBuilder:

    def __init__(self, enable, xattrs):
        self.xattrs = Object()
        self.xattrs._classname = 'ExtendedAttributes'  # pylint: disable=protected-access
        self.xattrs.enable = enable
        self.xattrs.attributes = xattrs

    @staticmethod
    def default():
        return ExtendedAttributesBuilder(True, [ExtendedAttribute(ExtendedAttributes.MacOS, True)])

    @staticmethod
    def disabled():
        return ExtendedAttributesBuilder(False, [ExtendedAttribute(ExtendedAttributes.MacOS, True)])

    def add(self, xattr):
        """
        :param cterasdk.core.enum.ExtendedAttributes xattr: Extended attribute
        """
        self.xattrs.append(xattr)

    def build(self):
        return self.xattrs


class RoleSettings(Object):  # pylint: disable=too-many-instance-attributes
    """
    Role Settings

    :ivar str name: Role name
    :ivar bool sudo: Super user
    :ivar bool enable_remote_wipe: Allow Remote Wipe for Devices
    :ivar bool enable_sso: Super user
    :ivar bool enable_seeding_export: Allow Seeding Export
    :ivar bool enable_seeding_import: Allow Seeding Import
    :ivar bool access_end_user_folders: Access End User Folders
    :ivar bool update_settings: Modify Virtual Portal Settings
    :ivar bool update_roles: Modify Roles
    :ivar bool update_account_emails: Modify User Email
    :ivar bool update_account_password: Modify User Password
    :ivar bool manage_cloud_drives: Manage Cloud Folders
    :ivar bool manage_plans: Manage Plans
    :ivar bool manage_logs: Manage Log Settings
    :ivar bool allow_folders_files_permanent_delete: Allow Folders Files Permanent Delete
    :ivar bool can_manage_legal_holds: Manage Legal Holds
    :ivar bool can_manage_compliance_settings: Manage Compliance Settings
    """
    # pylint: disable=too-many-arguments, too-many-locals
    def __init__(self, name, sudo, enable_remote_wipe, enable_sso, enable_seeding_export, enable_seeding_import, access_end_user_folders,
                 update_settings, update_roles, update_account_emails, update_account_password, manage_cloud_drives, manage_plans,
                 manage_users, manage_logs, allow_folders_files_permanent_delete, can_manage_legal_holds, can_manage_compliance_settings):
        super().__init__()
        self.name = name
        self.sudo = sudo
        self.enable_remote_wipe = enable_remote_wipe
        self.enable_sso = enable_sso
        self.enable_seeding_export = enable_seeding_export
        self.enable_seeding_import = enable_seeding_import
        self.access_end_user_folders = access_end_user_folders
        self.update_settings = update_settings
        self.update_roles = update_roles
        self.update_account_emails = update_account_emails
        self.update_account_password = update_account_password
        self.manage_cloud_drives = manage_cloud_drives
        self.manage_plans = manage_plans
        self.manage_users = manage_users
        self.manage_logs = manage_logs
        self.allow_folders_files_permanent_delete = allow_folders_files_permanent_delete
        self.can_manage_legal_holds = can_manage_legal_holds
        self.can_manage_compliance_settings = can_manage_compliance_settings

    def to_server_object(self):
        param = Object()
        param._classname = 'PortalRoleSettings'  # pylint: disable=protected-access
        param.name = self.name
        param.superUser = self.sudo
        param.allowRemoteWipe = self.enable_remote_wipe
        param.allowSSO = self.enable_sso
        param.allowSeedingExport = self.enable_seeding_export
        param.allowSeedingImport = self.enable_seeding_import
        param.canAccessEndUserFolders = self.access_end_user_folders
        param.canChangePortalSettings = self.update_settings
        param.canChangeRoles = self.update_roles
        param.canChangeUserEmail = self.update_account_emails
        param.canChangeUserPassword = self.update_account_password
        param.canManageAllFolders = self.manage_cloud_drives
        param.canManagePlans = self.manage_plans
        param.canManageUsers = self.manage_users
        param.canManageLogSettings = self.manage_logs
        param.allowFoldersFilesPermanentDelete = self.allow_folders_files_permanent_delete
        param.canManageLegalHolds = self.can_manage_legal_holds
        param.canManageComplianceSetting = self.can_manage_compliance_settings
        return param

    @staticmethod
    def from_server_object(server_object):
        params = {
            'name': server_object.name,
            'sudo': server_object.superUser,
            'enable_remote_wipe': server_object.allowRemoteWipe,
            'enable_sso': server_object.allowSSO,
            'enable_seeding_export': server_object.allowSeedingExport,
            'enable_seeding_import': server_object.allowSeedingImport,
            'access_end_user_folders': server_object.canAccessEndUserFolders,
            'update_settings': server_object.canChangePortalSettings,
            'update_roles': server_object.canChangeRoles,
            'update_account_emails': server_object.canChangeUserEmail,
            'update_account_password': server_object.canChangeUserPassword,
            'manage_cloud_drives': server_object.canManageAllFolders,
            'manage_plans': server_object.canManagePlans,
            'manage_users': server_object.canManageUsers,
            'manage_logs': server_object.canManageLogSettings,
            'allow_folders_files_permanent_delete': server_object.allowFoldersFilesPermanentDelete,
            'can_manage_legal_holds': server_object.canManageLegalHolds,
            'can_manage_compliance_settings': server_object.canManageComplianceSetting
        }
        return RoleSettings(**params)



import queue
import json
import copy
from xml.etree.ElementTree import Element, SubElement, tostring
from xml.dom import minidom

from .types import XMLTypes
from ..common import Item, Object, Device


_sdk_hidden = [
    'password',
    'awsSecretKey',
    'sharedSecret',
    'passPhraseSalt',
    'encPassphrase',
    'encryptedFolderKey',
    'oldPassword',
    'newPassword',
    'secretkey',
    'activationCode',
    'masterPassword',
    'masterKey',
    'secretAccess'
]


__protected__ = '*** Protected Value ***'


def _to_protected_dict(o):
    ret = copy.deepcopy(o.__dict__)
    for key in _sdk_hidden:
        if key in ret:
            ret[key] = __protected__
    return ret


def tojsonstr(obj, pretty_print=True, no_log=True):
    """
    Convert a Python object to a JSON string.

    :param object obj: the Python object
    :param bool pretty_print: Whether to format the JSON string, defaults to ``True``
    :param book no_log: Hide sensitive values in the log messages
    :return: JSON string of the object
    :rtype: str
    """
    indent = 5 if pretty_print else None
    if no_log:
        return json.dumps(obj, default=_to_protected_dict, indent=indent)
    return json.dumps(obj, default=lambda o: o.__dict__, indent=indent)


def toxmlstr(obj, pretty_print=False, no_log=False):
    """
    Convert a Python object to an XML string

    :param object obj: the Python object
    :param bool pretty_print: whether to format the XML string, defaults to ``False``
    :return: XML string of the object
    :rtype: str
    """
    if obj is None:
        return None
    xml = toxml(obj, no_log)
    if pretty_print:
        string = minidom.parseString(tostring(xml)).toprettyxml(indent="   ")
        return ''.join(string.split('\n', 1)[1:])
    return tostring(xml, 'utf-8')


def toxml(obj, no_log=False):  # pylint: disable=too-many-branches
    root = Item()
    root.node = None
    root.parent = None
    root.obj = obj

    q = queue.Queue()
    q.put(root)
    while not q.empty():
        item = q.get()
        if isinstance(item.obj, (str, int, float, complex, bool)):
            item.node = CreateElement(item.parent, XMLTypes.VAL)
            if isinstance(item.obj, bool):
                item.node.text = str(item.obj).lower()
            else:
                if no_log is True and item.parent and item.parent.get(XMLTypes.ID) in _sdk_hidden:
                    item.node.text = __protected__
                else:
                    item.node.text = str(item.obj)
        elif isinstance(item.obj, list):
            item.node = CreateElement(item.parent, XMLTypes.LIST)
            for member in item.obj:
                kid = Item()
                kid.node = None
                kid.parent = item.node
                kid.obj = member
                q.put(kid)
        elif isinstance(item.obj, Device):  # db.xml
            item.node = CreateElement(item.parent, XMLTypes.DB)
            for key, value in [
                (XMLTypes.NS, item.obj.namespace),
                (XMLTypes.LOCATION, item.obj.location),
                (XMLTypes.ID, item.obj.id),
                (XMLTypes.VERSION, item.obj.version),
                (XMLTypes.FIRMWARE, item.obj.firmware)
            ]:
                if value is not None:
                    item.node.set(key, value)

            kid = Item()
            kid.node = None
            kid.parent = item.node
            kid.obj = item.obj.config
            q.put(kid)
        elif isinstance(item.obj, Object):
            item.node = CreateElement(item.parent, XMLTypes.OBJ)
            classname = item.obj.__dict__.get('_classname')  # Convert { "_classname" : "ShareConfig" }
            if classname is not None:
                item.node.set(XMLTypes.CLASS, classname)
            uuid = item.obj.__dict__.get('_uuid')  # Convert { "_uuid" : "6f0e8c79-..." }
            if uuid is not None:
                item.node.set(XMLTypes.UUID, uuid)
            for attribute_name in item.obj.__dict__:
                if attribute_name.startswith('_'):
                    continue
                att = SubElement(item.node, XMLTypes.ATT)
                att.set(XMLTypes.ID, attribute_name)
                kid = Item()
                kid.node = None
                kid.parent = att
                kid.obj = item.obj.__dict__[attribute_name]
                q.put(kid)

    return root.node


def CreateElement(parent, tag):
    if parent is not None:
        element = SubElement(parent, tag)
    else:
        element = Element(tag)
    return element



import copy
import base64
from ..common import Object, utils


class ByteRange:

    def __init__(self, start=None, end=None):
        """
        Initialize a Byte Range.

        :param int,optional start: Start byte, defaults to 0
        :param int,optional end: End byte, defaults to EOF
        """
        self._eof = True
        self._start = start if start is not None and start > -1 else 0
        if end is not None:
            if end < self._start:
                raise ValueError(f'Byte range end {end} must be greater than or equal to range start {start}.')
            self._eof = False
            self._end = end

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def eof(self):
        return self._eof

    @eof.setter
    def eof(self, eof):
        self._eof = eof

    @staticmethod
    def default():
        return ByteRange(0)


class CompressionLib:
    """
    Compression Library

    :ivar str Snappy: Snappy
    :ivar int Gzip: Gzip
    :ivar int Off: No Compression
    """
    Snappy = 'SNAPPY'
    Gzip = 'GZIP'
    Off = 'NONE'


class Chunk(Object):

    def __init__(self, index, offset, url, length):
        """
        Initialize a Chunk.

        :param int index: Chunk index.
        :param int offset: Chunk offset.
        :param str url: Signed URL.
        :param int length: Object length.
        """
        super().__init__(
            index=index,
            offset=offset,
            url=url,
            length=length
        )


class Metadata(Object):
    """
    CTERA Direct IO File Metadata
    """

    def __init__(self, file_id, server_object):
        """
        Initialize a Direct IO metadata response object.

        :param int file_id: File ID.
        :param cterasdk.common.object.Object server_object: Response Object.
        """
        super().__init__(
            file_id=file_id,
            encrypted=server_object.encrypt_info.data_encrypted,
            compressed=server_object.compression_type != CompressionLib.Off,
            chunks=Metadata._format_chunks(server_object.chunks)
        )
        self.encryption_key = server_object.encrypt_info.wrapped_key if self.encrypted else None
        self.compression_library = server_object.compression_type if self.compressed else None
        last_chunk = self.chunks[-1]
        self.size = last_chunk.offset + last_chunk.length

    @staticmethod
    def _format_chunks(server_object):
        """
        Create Chunks.

        :param int file_id: File ID.
        :param cterasdk.common.object.Object server_object: Server response.
        :param list[int] blocks: List of block numbers to retrieve.
        :returns: Chunk objects
        :rtype: list[cterasdk.direct.types.Chunk]
        """
        offset = 0
        chunks = []
        for index, chunk in enumerate(server_object, 1):
            chunks.append(Chunk(index, offset, chunk.url, chunk.len))
            offset = offset + chunk.len
        return chunks

    def serialize(self):
        """
        Serialize Direct IO metadata to a dictionary.
        """
        x = copy.deepcopy(self)
        if self.encrypted:
            x.encryption_key = utils.utf8_decode(base64.b64encode(self.encryption_key))
        return x


class Block:
    """Block"""

    def __init__(self, file_id, number, offset, data, length):
        """
        Initialize a Block.

        :param int file_id: File ID.
        :param int number: Block number.
        :param int offset: Block offset.
        :param bytes data: Bytes
        :param int length: Block length.
        """
        self._file_id = file_id
        self._number = number
        self._offset = offset
        self._data = data
        self._length = length

    @property
    def file_id(self):
        return self._file_id

    @property
    def number(self):
        return self._number

    @property
    def offset(self):
        return self._offset

    @property
    def data(self):
        return self._data

    @property
    def length(self):
        return self._length

    def fragment(self, byte_range):
        if byte_range.start > self._offset and byte_range.start < self._offset + self._length:
            start = byte_range.start - self._offset
        else:
            start = None

        if not byte_range.eof and byte_range.end < self._offset + self._length:
            end = byte_range.end - self._offset + 1
        else:
            end = None

        data = self._data[start:end]
        return Block(self._file_id, self._number, self._offset + start if start else self._offset, data, len(data))
